"""
Custom OpenMM pair potentials
=============================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains implementations of commonly used pair potentials
that are not available in OpenMM, such as the Gaussian and Yukawa
potentials. Generally, the pair potentials are named after their LAMMPS
:code:`pair_style` counterparts, if available.
"""

from typing import Union

import numpy as np
import openmm
from openmm import unit

from .unit import VACUUM_PERMITTIVITY


def _setup_pair(
    cnbforce: openmm.CustomNonbondedForce,
    cutoff: Union[float, unit.Quantity],
    global_params: dict[str, Union[float, unit.Quantity]],
    per_params: list[str],
    tab_funcs: dict[
        str,
        Union[
            np.ndarray[Union[int, float]],
            openmm.unit.Quantity,
            openmm.Discrete2DFunction,
        ],
    ],
    method: int = openmm.CustomNonbondedForce.CutoffPeriodic,
) -> None:
    r"""
    Sets up a :class:`openmm.CustomNonbondedForce` object.

    Parameters
    ----------
    cnbforce : `openmm.CustomNonbondedForce`
        Custom nonbonded force object.

    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance :math:`r_\mathrm{cut}` for all nonbonded
        interactions in the simulation sytem. Must be less than half the
        periodic simulation box dimensions.

        **Reference unit**: :math:`\mathrm{nm}`.

    global_params : `dict`, optional
        Global parameters.

    per_params : `list`, optional
        Per-particle parameters.

    tab_funcs : `dict`, optional
        Tabulated functions.

    method : `int`, default: :code:`2`
        Cutoff method.
    """

    if global_params:
        for param in global_params.items():
            cnbforce.addGlobalParameter(*param)
    if per_params:
        for param in per_params:
            cnbforce.addPerParticleParameter(param)
    if tab_funcs:
        for name, func in tab_funcs.items():
            if not isinstance(func, openmm.Discrete2DFunction):
                func = openmm.Discrete2DFunction(*func.shape, func.ravel().tolist())
            cnbforce.addTabulatedFunction(name, func)
    cnbforce.setCutoffDistance(cutoff)
    cnbforce.setNonbondedMethod(method)


def coul_gauss(
    cutoff: Union[float, unit.Quantity],
    tol: float = 1e-4,
    *,
    g_ewald: Union[float, unit.Quantity] = None,
    dims: Union[np.ndarray[float], unit.Quantity] = None,
    mix: str = "default",
    per_params: list = None,
    global_params: dict[str, Union[float, unit.Quantity]] = None,
    tab_funcs: dict[
        str,
        Union[
            np.ndarray[Union[int, float]],
            openmm.unit.Quantity,
            openmm.Discrete2DFunction,
        ],
    ] = None,
) -> tuple[openmm.CustomNonbondedForce, openmm.NonbondedForce]:
    r"""
    Implements the smeared Coulomb pair potential.

    The charges have spherical Gaussian-distributed charge
    distributions [1]_ [2]_

    .. math::

       u_\mathrm{elec}(r_{12})=\frac{q_1q_2}{4\pi\varepsilon_0r_{12}}
       \mathrm{erf}(\alpha_{12}r_{12})

    where :math:`q_1` and :math:`q_2` are the particle charges in
    :math:`\mathrm{e}`, :math:`\varepsilon_0` is the vacuum
    permittivity, and :math:`\alpha_{12}=\sqrt{\alpha_1^2+\alpha_2^2}`
    is an inverse screening length or damping parameter in
    :math:`\mathrm{nm}^{-1}`. Effectively, electrostatic interactions
    are dampened for charges at small separation distances but remain
    unchanged at large separation distances.

    Additionally, an implementation of the electrostatic pair potential
    in the Gaussian core model

    .. math::

       u_\mathrm{elec}(r_{12})=\frac{q_1q_2}{4\pi\varepsilon_0r_{12}}
       \mathrm{erf}\left(\frac{\pi^{1/2}}{2^{1/2}a_{12}}r_{12}\right)

    is available, where :math:`a_{12}=\sqrt{a_1^2+a_2^2}` with
    :math:`a_1` and :math:`a_2` being the electrostatic smearing radii
    in :math:`\mathrm{nm}`.

    To account for the solvent polarization implicitly using its
    relative permittivity :math:`\varepsilon_\mathrm{r}`, scale the
    particle charges :math:`q_1` and :math:`q_2` by
    :math:`1/\sqrt{\varepsilon_\mathrm{r}}`.

    After creating the pair potentials, particles should be registered
    using :meth:`openmm.openmm.NonbondedForce.addParticle` and
    :meth:`openmm.openmm.CustomNonbondedForce.addParticle`.

    Parameters
    ----------
    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance for all nonbonded interactions in the
        simulation sytem. Must be less than half the minimum periodic
        simulation box dimension.

        **Reference unit**: :math:`\mathrm{nm}`.

    tol : `float`, default: :code:`1e-4`
        Error tolerance :math:`\delta` for Ewald summation.

    g_ewald : `float` or `openmm.unit.Quantity`, keyword-only, optional
        Ewald splitting parameter :math:`g_\mathrm{Ewald}`. If not
        provided, it is computed using

        .. math::

           g_\mathrm{Ewald}=\frac{\sqrt{-\log(2\delta)}}{r_\mathrm{cut}}

        **Reference unit**: :math:`\mathrm{nm}^{-1}`.

    dims : `array_like` or `openmm.unit.Quantity`, keyword-only, optional
        System dimensions. Must be provided with `g_ewald` to
        calculate the number of mesh nodes :math:`n_\mathrm{mesh}`.
        Both `dims` and `g_ewald` must either have units or be unitless.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    mix : `str`, keyword-only, default: :code:`"default"`
        Mixing rule for :math:`\alpha_{12}`.

        .. container::

           **Valid values**:

           * :code:`"default"`: Default mixing rule.

             .. math::

                \alpha_{12}=\frac{\alpha_1\alpha_2}
                {\sqrt{\alpha_1^2+\alpha_2^2}}

             **Per-particle parameters**:

             * :class:`openmm.openmm.CustomNonbondedForce`:
               :math:`(q_i,\,\alpha_i)`.
             * :class:`openmm.openmm.NonbondedForce`:
               :math:`(q_i,\,\sigma_i,\,\epsilon_i)`.

           * :code:`"core"`: Gaussian core model.

             .. math::

                \begin{gather*}
                  a_{12}^2=a_1^2+a_2^2\\
                  \alpha_{12}=\sqrt{\frac{\pi}{2a_{12}^2}}
                \end{gather*}

             This is equivalent to setting
             :math:`\alpha_i=\sqrt{\pi/(2a_i^2)}` in the default mixing
             rule.

             * :class:`openmm.openmm.CustomNonbondedForce`:
               :math:`(q_i,\,a_i)`.
             * :class:`openmm.openmm.NonbondedForce`:
               :math:`(q_i,\,\sigma_i,\,\epsilon_i)`.

           * :code:`"alpha12 = ...;"`: Custom mixing rule. The string
             containing the expression for :math:`\alpha_{12}` must be
             written in valid C++ syntax, with any custom global and
             per-particle parameters and tabulated functions defined in
             `global_params`, `per_params`, and `tab_funcs`,
             respectively.

             **Per-particle parameters**:

             * :class:`openmm.openmm.CustomNonbondedForce`:
               :math:`(q_i,\,\ldots)`.
             * :class:`openmm.openmm.NonbondedForce`:
               :math:`(q_i,\,\sigma_i,\,\epsilon_i)`.

        .. tip::

           To disable the Lennard-Jones potential, set
           :math:`\sigma_i=0\,\mathrm{nm}` and
           :math:`\epsilon_i=0\,\mathrm{kJ/mol}` for all particles.

    global_params : `dict`, keyword-only, optional
        Additional global parameters for use in the definition of
        :math:`\alpha_{12}`.

    per_params : `list`, keyword-only, optional
        Additional per-particle parameters for use in the definition of
        :math:`\alpha_{12}`.

    tab_funcs : `dict`, keyword-only, optional
        Optional tabulated functions for use in the definition of
        :math:`\alpha_{12}`.

    Returns
    -------
    pair_coul_gauss_dir : `openmm.CustomNonbondedForce`
        Short-range electrostatic contribution evaluated in real space.

    pair_coul_gauss_rec : `openmm.NonbondedForce`
        Long-range electrostatic contribution evaluated in Fourier
        (reciprocal) space.

    References
    ----------
    .. [1] Kiss, P. T.; Sega, M.; Baranyai, A. Efficient Handling of
       Gaussian Charge Distributions: An Application to Polarizable
       Molecular Models. *J. Chem. Theory Comput.* **2014**, *10* (12),
       5513–5519. https://doi.org/10.1021/ct5009069.

    .. [2] Eslami, H.; Khani, M.; Müller-Plathe, F. Gaussian Charge
       Distributions for Incorporation of Electrostatic Interactions in
       Dissipative Particle Dynamics: Application to Self-Assembly of
       Surfactants. *J. Chem. Theory Comput.* **2019**, *15* (7),
       4197–4207. https://doi.org/10.1021/acs.jctc.9b00174.
    """

    if g_ewald is None:
        g_ewald = np.sqrt(-np.log(2 * tol)) / cutoff

    if global_params is None:
        global_params = {}
    global_params |= {
        "G_EWALD": g_ewald,
        "ONE_4PI_EPS0": unit.AVOGADRO_CONSTANT_NA / (4 * np.pi * VACUUM_PERMITTIVITY),
    }
    if mix == "default":
        mix = "alpha12=alpha1*alpha2/sqrt(alpha1^2+alpha2^2);"
        per_params = ["alpha"]
    elif mix == "core":
        mix = f"alpha12=sqrt({np.pi}/(2*(a1^2+a2^2)));"
        per_params = ["a"]

    pair_coul_gauss_dir = openmm.CustomNonbondedForce(
        f"ONE_4PI_EPS0*q1*q2*(erf(alpha12*r)-erf(G_EWALD*r))/r;{mix}"
    )
    pair_coul_gauss_dir.addPerParticleParameter("q")
    _setup_pair(pair_coul_gauss_dir, cutoff, global_params, per_params, tab_funcs)
    pair_coul_gauss_rec = lj_coul(cutoff, tol, g_ewald=g_ewald, dims=dims)
    pair_coul_gauss_rec.setIncludeDirectSpace(False)
    return pair_coul_gauss_dir, pair_coul_gauss_rec


def dpd(
    cutoff: Union[float, unit.Quantity],
    cutoff_dpd: Union[float, unit.Quantity] = None,
    *,
    mix: str = None,
    per_params: list = None,
    global_params: dict[str, Union[float, unit.Quantity]] = None,
    tab_funcs: dict[
        str,
        Union[np.ndarray[Union[int, float]], unit.Quantity, openmm.Discrete2DFunction],
    ] = None,
) -> openmm.CustomNonbondedForce:
    r"""
    Implements the conservative part of the dissipative particle
    dynamics (DPD) force.

    .. note::

       This does not include an implementation of the DPD thermostat.

    The potential energy between two DPD beads is given by:

    .. math::

       u_\mathrm{DPD}(r_{12})=\frac{1}{2}A_{12}r_\mathrm{cut}
       \left(1-\frac{r}{r_\mathrm{cut}}\right)^2

    where :math:`A_{12}` is the conservative force parameter in
    :math:`\mathrm{kJ/(mol\cdot nm)}` and :math:`r_\mathrm{cut}` is the
    interaction cutoff distance in :math:`\mathrm{nm}`.

    As :math:`A_{12}` has no well-defined mixing rule, it must be

    * evaluated using a custom mixing rule in `mix` with necessary
      per-particle parameters in `per_params`,
    * specified as a global parameter in `global_params`, or
    * provided for each pair of atom types in `tab_funcs` while
      specifying :code:`mix="A12=A(type1,type2);"` and
      :code:`per_params=("type",)`.

    After creating the pair potential, particles should be registered
    using :meth:`openmm.openmm.CustomNonbondedForce.addParticle`.

    Parameters
    ----------
    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance for all nonbonded interactions in the
        simulation sytem. Must be less than half the minimum periodic
        simulation box dimension.

        **Reference unit**: :math:`\mathrm{nm}`.

    cutoff_dpd : `float` or `openmm.unit.Quantity`, optional
        Cutoff distance :math:`r_\mathrm{cut}` for the DPD potential.
        Must be less than the shared cutoff distance. If not provided,
        it is set to the shared cutoff distance.

        **Reference unit**: :math:`\mathrm{nm}`.

    mix : `str`, keyword-only, optional
        Mixing rule for :math:`A_{12}`. Must be defined if
        :math:`A_{12}` is not a global parameter. The string containing
        the expression for :math:`A_{12}` must be written in valid C++
        syntax, with any custom global and per-particle parameters and
        tabulated functions defined in `global_params`, `per_params`,
        and `tab_funcs`, respectively.

    global_params : `dict`, keyword-only, optional
        Additional global parameters for use in the definition of
        :math:`A_{12}`.

    per_params : `list`, keyword-only, optional
        Additional per-particle parameters for use in the definition of
        :math:`A_{12}`.

    tab_funcs : `dict`, keyword-only, optional
        Optional tabulated functions for use in the definition of
        :math:`A_{12}`.

    Returns
    -------
    pair_dpd : `openmm.CustomNonbondedForce`
        DPD pair potential.
    """

    if isinstance(cutoff, unit.Quantity):
        cutoff = cutoff.value_in_unit(unit.nanometer)
    if cutoff_dpd is None:
        cutoff_dpd = cutoff
    else:
        if isinstance(cutoff_dpd, unit.Quantity):
            cutoff_dpd = cutoff_dpd.value_in_unit(unit.nanometer)
        if cutoff_dpd > cutoff:
            emsg = (
                "The cutoff distance for the dissipative particle "
                "dynamics (DPD) potential must be less than the "
                "shared cutoff distance."
            )
            raise ValueError(emsg)

    energy = f"0.5*A12*{cutoff_dpd}*(1-r/{cutoff_dpd})^2;"
    if mix:
        energy = f"{energy}{mix}"

    pair_dpd = openmm.CustomNonbondedForce(energy)
    _setup_pair(pair_dpd, cutoff, global_params, per_params, tab_funcs)
    return pair_dpd


def gauss(
    cutoff: Union[float, unit.Quantity],
    cutoff_gauss: Union[float, unit.Quantity] = None,
    *,
    shift: bool = True,
    mix: str = "geometric",
    global_params: dict[str, Union[float, unit.Quantity]] = None,
    per_params: list = None,
    tab_funcs: dict[
        str,
        Union[np.ndarray[Union[int, float]], unit.Quantity, openmm.Discrete2DFunction],
    ] = None,
) -> openmm.CustomNonbondedForce:
    r"""
    Implements the generalized intermolecular Gaussian pair potential.

    The potential energy between two Gaussian particles is given by

    .. math::

       u_\mathrm{Gauss}(r_{12})=\alpha_{12}\exp{(-\beta_{12}r_{12}^2)}

    where :math:`\alpha_{12}` is a magnitude scaling constant in
    :math:`\mathrm{kJ/mol}` and :math:`\beta_{12}` is an exponential 
    prefactor in :math:`\mathrm{nm}^{-2}`.

    In addition to arithmetic and geometric mixing rules for
    :math:`\alpha_{12}` and :math:`\beta_{12}`, an implementation of the
    excluded volume pair potential in the Gaussian core model

    .. math::

        u_\mathrm{Gauss}(r_{12})=A\left(\frac{3}{2\pi\sigma_{12}^2}
        \right)^{3/2}\exp{\left(-\frac{3}{2\sigma_{12}^2}r_{12}^2\right)}

    is available, where :math:`A` is the repulsion parameter in
    :math:`\mathrm{nm}^3\cdot\mathrm{kJ/mol}` and :math:`\sigma_{12}` is
    the average particle size in :math:`\mathrm{nm}`.

    After creating the pair potential, particles should be registered
    using :meth:`openmm.openmm.CustomNonbondedForce.addParticle`.

    Parameters
    ----------
    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance for all nonbonded interactions in the
        simulation sytem. Must be less than half the minimum periodic
        simulation box dimension.

        **Reference unit**: :math:`\mathrm{nm}`.

    cutoff_gauss : `float` or `openmm.unit.Quantity`, optional
        Cutoff distance :math:`r_\mathrm{cut}` for the Gaussian
        potential. Must be less than the shared cutoff distance. If not
        provided, it is set to the shared cutoff distance.

        **Reference unit**: :math:`\mathrm{nm}`.

    shift : `bool`, keyword-only, default: :code:`True`
        Determines whether the Gaussian potential is shifted at its
        cutoff to :math:`0\,\mathrm{kJ/mol}`.

    mix : `str`, keyword-only, default: :code:`"geometric"`
        Mixing rule for :math:`\alpha_{12}` and :math:`\beta_{12}`.

        .. container::

           **Valid values**:

           * :code:`"arithmetic"`: Arithmetic combining rule.

             .. math::

                \begin{gather*}
                  \alpha_{12}=\sqrt{\alpha_1\alpha_2}\\
                  \beta_{12}=\frac{2}{\beta_1^{-1}+\beta_2^{-1}}
                \end{gather*}

             **Per-particle parameters**: :math:`(\alpha_i,\,\beta_i)`.

           * :code:`"geometric"`: Geometric mixing rule.

             .. math::

                \begin{gather*}
                  \alpha_{12}=\sqrt{\alpha_1\alpha_2}\\
                  \beta_{12}=\sqrt{\beta_1\beta_2}
                \end{gather*}

             **Per-particle parameters**: :math:`(\alpha_i,\,\beta_i)`.

           * :code:`"core"`: Gaussian core model.

             .. math::

                \begin{gather*}
                  \sigma_{12}^2=\sigma_1^2+\sigma_2^2\\
                  \beta_{12}=\frac{3}{2\sigma_{12}^2}\\
                  \alpha_{12}=A\left(
                  \frac{\beta_{12}}{\pi}\right)^{3/2}
                \end{gather*}

             **Global parameter**: :math:`A`.

             **Per-particle parameter**: :math:`\sigma_i`.

           * :code:`"alpha12 = ...; beta12 = ...;"`: Custom mixing rule.
             The string containing the expressions for
             :math:`\alpha_{12}` and :math:`\beta_{12}` must be written
             in valid C++ syntax, with any custom global and
             per-particle parameters and tabulated functions defined in
             `global_params`, `per_params`, and `tab_funcs`,
             respectively.

    global_params : `dict`, keyword-only, optional
        Optional global parameters for use in the definition of
        :math:`\alpha_{12}` and :math:`\beta_{12}`.

    per_params : `list`, keyword-only, optional
        Optional per-particle parameters for use in the definition of
        :math:`\alpha_{12}` and :math:`\beta_{12}`.

    tab_funcs : `dict`, keyword-only, optional
        Optional tabulated functions for use in the definition of
        :math:`\alpha_{12}` and :math:`\beta_{12}`.

    Returns
    -------
    pair_gauss : `openmm.CustomNonbondedForce`
        Gaussian pair potential.
    """

    if isinstance(cutoff, unit.Quantity):
        cutoff = cutoff.value_in_unit(unit.nanometer)
    if cutoff_gauss is None:
        cutoff_gauss = cutoff
    else:
        if isinstance(cutoff_gauss, unit.Quantity):
            cutoff_gauss = cutoff_gauss.value_in_unit(unit.nanometer)
        if cutoff_gauss > cutoff:
            emsg = (
                "The cutoff distance for the Gaussian potential "
                "must be less than the shared cutoff distance."
            )
            raise ValueError(emsg)

    prefix = f"step({cutoff_gauss}-r)*(" if cutoff != cutoff_gauss else "("
    root = "alpha12*exp(-beta12*r^2)"
    suffix = f"-ucut);ucut=alpha12*exp(-beta12*{cutoff_gauss}^2);" if shift else ");"
    if mix == "arithmetic":
        mix = "alpha12=sqrt(alpha1*alpha2);beta12=2/(1/beta1+1/beta2);"
        per_params = ["alpha", "beta"]
    elif mix == "geometric":
        mix = "alpha12=sqrt(alpha1*alpha2);beta12=sqrt(beta1*beta2);"
        per_params = ["alpha", "beta"]
    elif "core" in mix:
        mix = mix.replace(
            "core",
            f"alpha12=A*(beta12/{np.pi})^(3/2);beta12=3/(2*sigma12sq);"
            "sigma12sq=sigma1^2+sigma2^2",
        )
        if not mix.endswith(";"):
            mix += ";"
        if "A" not in mix and "A" not in global_params:
            raise ValueError("Global parameter 'A' not specified.")
        if per_params is None:
            per_params = []
        per_params += ["sigma"]

    pair_gauss = openmm.CustomNonbondedForce(f"{prefix}{root}{suffix}{mix}")
    _setup_pair(pair_gauss, cutoff, global_params, per_params, tab_funcs)
    return pair_gauss


def lj_coul(
    cutoff: Union[float, unit.Quantity],
    tol: float = 1e-4,
    *,
    g_ewald: Union[float, unit.Quantity] = None,
    dims: Union[float, unit.Quantity] = None,
) -> openmm.NonbondedForce:
    r"""
    Implements the standard Lennard-Jones (LJ) and Coulomb pair
    potentials.

    The potential energy between two charged LJ particles is given by

    .. math::

       \begin{gather*}
         u_\mathrm{Coul}(r_{12})=\frac{q_1q_2}{4\pi\varepsilon_0r_{12}}\\
         u_\mathrm{LJ}(r_{12})=4\epsilon_{12}\left[\left(
         \frac{\sigma_{12}}{r_{12}}\right)^{12}
         -\left(\frac{\sigma_{12}}{r_{12}} \right)^6\right]
       \end{gather*}

    where :math:`q_1` and :math:`q_2` are the particle charges in
    :math:`\mathrm{e}`, :math:`\varepsilon_0` is the vacuum
    permittivity, :math:`\sigma_{12}` is the averged particle size in
    :math:`\mathrm{nm}`, and :math:`\epsilon_{12}` is the dispersion
    energy in :math:`\mathrm{kJ/mol}`. :math:`q_1`, :math:`q_2`,
    :math:`\sigma_{12}`, and :math:`\epsilon_{12}` are determined from
    per-particle parameters `charge`, `sigma`, and `epsilon`,
    respectively, which are set in the main script using
    :meth:`openmm.openmm.NonbondedForce.addParticle`.

    The mixing rule for :math:`\sigma_{12}` and :math:`\epsilon_{12}` is
    the Lorentz–Berthelot combining rule:

    .. math::

       \begin{gather*}
         \sigma_{12}=\frac{\sigma_1+\sigma_2}{2}\\
         \epsilon_{12}=\sqrt{\epsilon_1\epsilon_2}
       \end{gather*}

    To disable the LJ potential and use a different excluded volume
    interaction potential but keep the Coulomb potential, set
    :math:`\sigma_{12}=0\,\mathrm{nm}` and
    :math:`\epsilon_{12}=0\,\mathrm{kJ/mol}` for all particles.

    To account for a solvent implicitly using its relative permittivity
    :math:`\varepsilon_\mathrm{r}`, scale the particle charges
    :math:`q_i` by :math:`1/\sqrt{\varepsilon_\mathrm{r}}`.

    Parameters
    ----------
    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance :math:`r_\mathrm{cut}` for all nonbonded
        interactions in the simulation sytem. Must be less than half the
        minimum periodic simulation box dimension.

        **Reference unit**: :math:`\mathrm{nm}`.

    tol : `float`, default: :code:`1e-4`
        Error tolerance :math:`\delta` for Ewald summation.

    g_ewald : `float` or `openmm.unit.Quantity`, keyword-only, optional
        Ewald splitting parameter :math:`g_\mathrm{Ewald}`. If not
        provided, it is computed using

        .. math::

           g_\mathrm{Ewald}=\frac{\sqrt{-\log(2\delta)}}{r_\mathrm{cut}}

        **Reference unit**: :math:`\mathrm{nm}^{-1}`.

    dims : `float` or `openmm.unit.Quantity`, keyword-only, optional
        System dimensions. Must be provided with `g_ewald` to
        calculate the number of mesh nodes :math:`n_\mathrm{mesh}`.
        Both `dims` and `g_ewald` must either have units or be unitless.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    Returns
    -------
    pair_lj_coul : `openmm.NonbondedForce`
        Hybrid LJ and Coulomb pair potentials.
    """

    pair_lj_coul = openmm.NonbondedForce()
    pair_lj_coul.setCutoffDistance(cutoff)
    pair_lj_coul.setNonbondedMethod(openmm.NonbondedForce.PME)
    if g_ewald is None or dims is None:
        pair_lj_coul.setEwaldErrorTolerance(tol)
    else:
        n_mesh = np.ceil(2 * g_ewald * dims / (3 * tol ** (1 / 5)))
        pair_lj_coul.setPMEParameters(g_ewald, *n_mesh)
    return pair_lj_coul


def ljts(
    cutoff: Union[float, unit.Quantity],
    cutoff_ljts: Union[float, unit.Quantity] = None,
    *,
    coefs: Union[dict[str, float], tuple[float, float, float]] = (1, 1, 4),
    powers: Union[dict[str, float], tuple[float, float]] = (12, 6),
    shift: bool = True,
    mix: str = "arithmetic",
    mie: bool = False,
    wca: bool = False,
    global_params: dict[str, Union[float, unit.Quantity]] = None,
    per_params: list = None,
    tab_funcs: dict[
        str,
        Union[np.ndarray[Union[int, float]], unit.Quantity, openmm.Discrete2DFunction],
    ] = None,
) -> openmm.CustomNonbondedForce:
    r"""
    Implements the generalized Lennard-Jones truncated and shifted
    (LJTS) pair potential and its derivatives, such as the Mie
    and Weeks–Chandler–Andersen (WCA) potentials.

    The generalized LJ potential is given by

    .. math::

       u_\mathrm{LJ}(r_{12})=C\epsilon_{12}
       \left[A\left(\frac{\sigma_{12}}{r_{12}}\right)^{\gamma_\mathrm{r}}
       -B\left(\frac{\sigma_{12}}{r_{12}}\right)^{\gamma_\mathrm{a}}\right]

    where :math:`\sigma_{12}` is the average particle size in
    :math:`\mathrm{nm}`, :math:`\epsilon_{12}` is the dispersion
    energy in :math:`\mathrm{kJ/mol}`, and :math:`A`, :math:`B`,
    :math:`C`, :math:`\gamma_\mathrm{r}`, and :math:`\gamma_\mathrm{a}`
    are constants.

    The standard 12-6 LJ potential, which is the most frequently used
    form, uses :math:`A=1`, :math:`B=1`, :math:`C=4`,
    :math:`\gamma_\mathrm{r}=12`, and :math:`\gamma_\mathrm{a}=6`.

    While :math:`\gamma_\mathrm{a}=6` is justified by the London
    dispersion force, no such rationale exists for
    :math:`\gamma_\mathrm{r}=12`. By relaxing the restrictions on what
    values :math:`\gamma_\mathrm{r}` and :math:`\gamma_\mathrm{a}` can
    take on and keeping :math:`A=1` and :math:`B=1`, the Mie potential
    is obtained:

    .. math::

       \begin{gather*}
         u_\mathrm{Mie}(r_{12})=C_\mathrm{Mie}\epsilon_{12}
         \left[\left(\frac{\sigma_{12}}{r_{12}}\right)^{\gamma_\mathrm{r}}
         -\left(\frac{\sigma_{12}}{r_{12}}\right)^{\gamma_\mathrm{a}}\right]\\
         C_\mathrm{Mie}=\frac{\gamma_\mathrm{r}}
         {\gamma_\mathrm{r}-\gamma_\mathrm{a}}
         \left(\frac{\gamma_\mathrm{r}}{\gamma_\mathrm{a}}\right)
         ^{\gamma_\mathrm{a}/(\gamma_\mathrm{r}-\gamma_\mathrm{a})}
       \end{gather*}

    In dense fluids, it is common to use a purely repulsive potential.
    The standard 12-6 LJ potential can be truncated and shifted at a
    cutoff of :math:`r_\mathrm{cut}=2^{1/6}\sigma_{12}` to get the WCA
    potential, which is widely employed as a reference potential in
    simulations of liquid systems. A generalized WCA potential can be
    acquired by truncating and shifting the Mie potential instead at a
    cutoff of :math:`r_\mathrm{WCA}=(\gamma_\mathrm{r}/\gamma_\mathrm{a})
    ^{1/(\gamma_\mathrm{r}-\gamma_\mathrm{a})}\sigma_{12}`:

    .. math::

       u_\mathrm{WCA}(r_{12})=\begin{cases}
         C_\mathrm{Mie}\epsilon_{12}\left[
         \left(\dfrac{\sigma_{12}}{r_{12}}\right)^{\gamma_\mathrm{r}}
         -\left(\dfrac{\sigma_{12}}{r_{12}}\right)^{\gamma_\mathrm{a}}
         \right]+\epsilon_{12},&\mathrm{if}\,r_{12}<r_\mathrm{WCA}\\
         0,&\mathrm{if}\,r_{12}\geq r_\mathrm{WCA}
       \end{cases}

    In the case where flexibility is allowed for all constants and the
    cutoff, the generalized LJTS potential is obtained:

    .. math::

       u_\mathrm{LJTS}(r_{12})=\begin{cases}
         u_\mathrm{LJ}(r_{12})-u_\mathrm{LJ}(r_\mathrm{cut}),
         &\mathrm{if}\,r_{12}<r_\mathrm{cut}\\
         0,&\mathrm{if}\,r_{12}\geq r_\mathrm{cut}
       \end{cases}

    After creating the pair potentials, particles should be registered
    using :meth:`openmm.openmm.CustomNonbondedForce.addParticle`.

    Parameters
    ----------
    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance for all nonbonded interactions in the
        simulation sytem. Must be less than half the minimum periodic
        simulation box dimension.

        **Reference unit**: :math:`\mathrm{nm}`.

    cutoff_ljts : `float` or `openmm.unit.Quantity`, optional
        Cutoff distance :math:`r_\mathrm{cut}` for the LJTS potential.
        Must be less than the shared cutoff distance. If not provided,
        it is set to the shared cutoff distance. If the WCA potential is
        used, the dynamic cutoff :math:`r_\mathrm{WCA}` is used instead.

        **Reference unit**: :math:`\mathrm{nm}`.

    coefs : `tuple` or `dict`, keyword-only, default: :code:`(1, 1, 4)`
        Coefficients :math:`A`, :math:`B`, and :math:`C`, in that order.
        If a `dict` is provided, use the keys :code:`"A"`, :code:`"B"`,
        and :code:`"C"`, respectively. If the Mie or WCA potential is
        used, the appropriate coefficients are used instead.

    powers : `tuple` or `dict`, keyword-only, default: :code:`(12, 6)`
        Powers :math:`\gamma_\mathrm{r}` and :math:`\gamma_\mathrm{a}`,
        in that order. If a `dict` is provided, use the keys :code:`"r"`
        and :code:`"a"`, respectively.

    shift : `bool`, keyword-only, default: :code:`True`
        Determines whether the LJTS potential is shifted at its cutoff
        to :math:`0\,\mathrm{kJ/mol}`. If the WCA potential is used, a
        dynamic shift of :math:`\epsilon_{12}` is used instead.

    mie : `bool`, keyword-only, default: :code:`False`
        Determines whether to use the Mie potential. If :code:`True`,
        `coefs` is superseded by the correct values for the Mie
        potential. Only one of `mie` and `wca` can be :code:`True`.

    wca : `bool`, keyword-only, default: :code:`False`
        Determines whether to use the WCA potential. If :code:`True`,
        `cutoff_ljts`, `coefs`, and `shift` are superseded by the
        correct values for the WCA potential. Only one of `mie` and
        `wca` can be :code:`True`.

    mix : `str`, keyword-only, default: :code:`"arithmetic"`
        Mixing rule for :math:`\sigma_{12}` and :math:`\epsilon_{12}`.

        .. container::

           **Valid values**:

           * :code:`"arithmetic"`: Lorentz-Berthelot combining rule.

             .. math::

                \begin{gather*}
                  \sigma_{12}=\frac{\sigma_1+\sigma_2}{2}\\
                  \epsilon_{12}=\sqrt{\epsilon_1\epsilon_2}
                \end{gather*}

             **Per-particle parameters**: :math:`(\sigma_i,\,\epsilon_i)`.

           * :code:`"geometric"`: Geometric mixing rule.

             .. math::

                \begin{gather*}
                  \sigma_{12}=\sqrt{\sigma_1\sigma_2}\\
                  \epsilon_{12}=\sqrt{\epsilon_1\epsilon_2}
                \end{gather*}

             **Per-particle parameters**: :math:`(\sigma_i,\,\epsilon_i)`.

           * :code:`"sixthpower"`: Sixth-power mixing rule.

             .. math::

                \begin{gather*}
                  \sigma_{12}=\left(\frac{\sigma_1^6
                  +\sigma_2^6}{2}\right)^{1/6}\\
                  \epsilon_{12}=\frac{2\sqrt{\epsilon_1\epsilon_2}
                  \sigma_1^3\sigma_2^3}{\sigma_1^6+\sigma_2^6}
                \end{gather*}

             **Per-particle parameters**: :math:`(\sigma_i,\,\epsilon_i)`.

           * :code:`"sigma12 = ...; epsilon12 = ...;"`: Custom mixing
             rule. The string containing the expression for
             :math:`\sigma_{12}` and :math:`\epsilon_{12}` must be
             written in valid C++ syntax, with any custom global and
             per-particle parameters and tabulated functions defined in
             `global_params`, `per_params`, and `tab_funcs`,
             respectively.

    global_params : `dict`, keyword-only, optional
        Additional global parameters for use in the definition of
        :math:`\sigma_{12}` and :math:`\epsilon_{12}`.

    per_params : `list`, keyword-only, optional
        Additional per-particle parameters for use in the definition of
        :math:`\sigma_{12}` and :math:`\epsilon_{12}`.

    tab_funcs : `dict`, keyword-only, optional
        Optional tabulated functions for use in the definition of
        :math:`\sigma_{12}` and :math:`\epsilon_{12}`.

    Returns
    -------
    pair_ljts : `openmm.CustomNonbondedForce`
        LJTS pair potential.
    """

    if isinstance(cutoff, unit.Quantity):
        cutoff = cutoff.value_in_unit(unit.nanometer)
    if cutoff_ljts is None:
        cutoff_ljts = cutoff
    else:
        if isinstance(cutoff_ljts, unit.Quantity):
            cutoff_ljts = cutoff_ljts.value_in_unit(unit.nanometer)
        if cutoff_ljts > cutoff:
            emsg = (
                "The cutoff distance for the LJTS potential must be "
                "less than the shared cutoff distance."
            )
            raise ValueError(emsg)

    if mie and wca:
        raise ValueError("Both 'mie' and 'wca' are set to True.")
    elif mie or wca:
        coef_mie = (
            powers[0]
            / (powers[0] - powers[1])
            * (powers[0] / powers[1]) ** (powers[1] / (powers[0] - powers[1]))
        )
    if isinstance(powers, dict):
        powers = (powers["r"], powers["a"])
    if wca:
        cutoff_wca = (powers[0] / powers[1]) ** (1 / (powers[0] - powers[1]))
        root = (
            f"{coef_mie}*epsilon12*((sigma12/r)^{powers[0]}"
            f"-(sigma12/r)^{powers[1]})"
        )
        prefix = f"step({cutoff_wca}*sigma12-r)*("
        suffix = "+epsilon12);"
    else:
        if mie:
            coefs = (1, 1, coef_mie)
        elif isinstance(coefs, dict):
            coefs = (coefs["A"], coefs["B"], coefs["C"])
        root = (
            f"{coefs[2]}*epsilon12*({coefs[0]}*(sigma12/r)^{powers[0]}"
            f"-{coefs[1]}*(sigma12/r)^{powers[1]})"
        )
        prefix = f"step({cutoff_ljts}-r)*(" if cutoff != cutoff_ljts else "("
        suffix = (
            (
                f"-ucut);ucut={coefs[2]}*epsilon12"
                f"*({coefs[0]}*(sigma12/{cutoff_ljts})^{powers[0]}"
                f"-{coefs[1]}*(sigma12/{cutoff_ljts})^{powers[1]});"
            )
            if shift
            else ");"
        )
    if mix == "arithmetic":
        mix = "sigma12=(sigma1+sigma2)/2;epsilon12=sqrt(epsilon1*epsilon2);"
        per_params = ["sigma", "epsilon"]
    elif mix == "geometric":
        mix = "sigma12=sqrt(sigma1*sigma2);epsilon12=sqrt(epsilon1*epsilon2);"
        per_params = ["sigma", "epsilon"]
    elif mix == "sixthpower":
        mix = (
            "sigma12=((sigma1^6+sigma2^6)/2)^(1/6);"
            "epsilon12=2*sqrt(epsilon1*epsilon2)*sigma1^3*sigma2^3"
            "/(sigma1^6+sigma2^6);"
        )
        per_params = ["sigma", "epsilon"]

    pair_ljts = openmm.CustomNonbondedForce(f"{prefix}{root}{suffix}{mix}")
    _setup_pair(pair_ljts, cutoff, global_params, per_params, tab_funcs)
    return pair_ljts


def solvation(
    cutoff: Union[float, unit.Quantity],
    cutoff_solvation: Union[float, unit.Quantity] = None,
    *,
    mix: str = "arithmetic",
    per_params: list = None,
    global_params: dict[str, Union[float, unit.Quantity]] = None,
    tab_funcs: dict[
        str,
        Union[
            np.ndarray[Union[int, float]],
            openmm.unit.Quantity,
            openmm.Discrete2DFunction,
        ],
    ] = None,
) -> openmm.CustomNonbondedForce:
    r"""
    Implements the solvation pair potential.

    The solvation energy between particles is given by

    .. math::

       u_\mathrm{solv}(r_{12})=-S_{12}\left[\left(
       \frac{\sigma_{12}}{r_{12}}\right)^4
       -\left(\frac{\sigma_{12}}{r_\mathrm{cut}}\right)^4\right]

    where :math:`\sigma_{12}` is the size of the particle in
    :math:`\mathrm{nm}` and :math:`S_{12}` is the solvation strength
    in :math:`\mathrm{kJ/mol}`.

    After creating the pair potential, particles should be registered
    using :meth:`openmm.openmm.CustomNonbondedForce.addParticle`.

    Parameters
    ----------
    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance for all nonbonded interactions in the
        simulation sytem. Must be less than half the minimum periodic
        simulation box dimension.

        **Reference unit**: :math:`\mathrm{nm}`.

    cutoff_solvation : `float` or `openmm.unit.Quantity`, optional
        Cutoff distance :math:`r_\mathrm{cut}` for the solvation
        potential. Must be less than the shared cutoff distance. If not
        provided, it is set to the shared cutoff distance.

        **Reference unit**: :math:`\mathrm{nm}`.

    mix : `str`, keyword-only, default: :code:`"arithmetic"`
        Mixing rule for :math:`\sigma_{12}` and :math:`S_{12}`.

        .. container::

           **Valid values**:

           * :code:`"arithmetic"`: Arithmetic mixing rule.

             .. math::

                \begin{gather*}
                  \sigma_{12}=\frac{\sigma_1+\sigma_2}{2}\\
                  S_{12}=\sqrt{S_1S_2}
                \end{gather*}

             **Per-particle parameters**: :math:`(\sigma_i, S_i)`.

           * :code:`"geometric"`: Geometric mixing rule.

             .. math::

                \begin{gather*}
                  \sigma_{12}=\sqrt{\sigma_1\sigma_2}\\
                  S_{12}=\sqrt{S_1S_2}
                \end{gather*}

             **Per-particle parameters**: :math:`(\sigma_i, S_i)`.

           * :code:`"sigma12 = ...; S12 = ...;"`: Custom mixing rule.
             The string containing the expressions for
             :math:`\sigma_{12}` and :math:`S_{12}` must be written
             in valid C++ syntax, with any custom global and
             per-particle parameters and tabulated functions defined in
             `global_params`, `per_params`, and `tab_funcs`,
             respectively.

    global_params : `dict`, keyword-only, optional
        Additional global parameters for use in the definition of
        :math:`\sigma_{12}` and :math:`S_{12}`.

    per_params : `list`, keyword-only, optional
        Additional per-particle parameters for use in the definition of
        :math:`\sigma_{12}` and :math:`S_{12}`.

    tab_funcs : `dict`, keyword-only, optional
        Optional tabulated functions for use in the definition of
        :math:`\sigma_{12}` and :math:`S_{12}`.

    Returns
    -------
    pair_solv : `openmm.CustomNonbondedForce`
        Solvation pair potential.
    """

    if isinstance(cutoff, unit.Quantity):
        cutoff = cutoff.value_in_unit(unit.nanometer)
    if cutoff_solvation is None:
        cutoff_solvation = cutoff
    else:
        if isinstance(cutoff_solvation, unit.Quantity):
            cutoff_solvation = cutoff_solvation.value_in_unit(unit.nanometer)
        if cutoff_solvation > cutoff:
            emsg = (
                "The cutoff distance for the solvation potential "
                "must be less than the shared cutoff distance."
            )
            raise ValueError(emsg)

    root = f"-S12*((sigma12/r)^4-(sigma12/{cutoff_solvation})^4)"
    if mix == "arithmetic":
        mix = "sigma12=(sigma1+sigma2)/2;S12=sqrt(S1*S2);"
        per_params = ["sigma", "S"]
    elif mix == "geometric":
        mix = "sigma12=sqrt(sigma1*sigma2);S12=sqrt(S1*S2);"
        per_params = ["sigma", "S"]

    pair_solv = openmm.CustomNonbondedForce(f"{root}{mix}")
    _setup_pair(pair_solv, cutoff, global_params, per_params, tab_funcs)
    return pair_solv


def wca(
    cutoff: Union[float, unit.Quantity],
    *,
    mix: str = "arithmetic",
    powers: Union[dict[str, float], tuple[float, float]] = (12, 6),
    global_params: dict[str, Union[float, unit.Quantity]] = None,
    per_params: list = None,
    tab_funcs: dict[
        str,
        Union[np.ndarray[Union[int, float]], unit.Quantity, openmm.Discrete2DFunction],
    ] = None,
) -> openmm.CustomNonbondedForce:
    r"""
    Implements the generalized Weeks–Chander–Andersen (WCA) pair
    potential.

    The potential energy between two WCA particles is given by

    .. math::

       \begin{gather*}
         u_\mathrm{WCA}(r_{12})=\begin{cases}
           C_\mathrm{Mie}\epsilon_{12}\left[
           \left(\dfrac{\sigma_{12}}{r_{12}}\right)^{\gamma_\mathrm{r}}
           -\left(\dfrac{\sigma_{12}}{r_{12}}\right)^{\gamma_\mathrm{a}}
           \right]+\epsilon_{12},&\mathrm{if}\,r_{12}<r_\mathrm{WCA}\\
           0,&\mathrm{if}\,r_{12}\geq r_\mathrm{WCA}
         \end{cases}\\
         C_\mathrm{Mie}=\frac{\gamma_\mathrm{r}}
         {\gamma_\mathrm{r}-\gamma_\mathrm{a}}
         \left(\frac{\gamma_\mathrm{r}}{\gamma_\mathrm{a}}\right)
         ^{\gamma_\mathrm{a}/(\gamma_\mathrm{r}-\gamma_\mathrm{a})}
       \end{gather*}

    where :math:`\sigma_{12}` is the average particle size in
    :math:`\mathrm{nm}`, :math:`\epsilon_{12}` is the dispersion energy
    in :math:`\mathrm{kJ/mol}`, :math:`\gamma_\mathrm{r}` and
    :math:`\gamma_\mathrm{a}` are the powers for the repulsive and
    attractive terms, respectively, and
    :math:`r_\mathrm{WCA}=(\gamma_\mathrm{r}/\gamma_\mathrm{a})
    ^{1/(\gamma_\mathrm{r}-\gamma_\mathrm{a})}\sigma_{12}` is the
    cutoff distance.

    After creating the pair potentials, particles should be registered
    using :meth:`openmm.openmm.CustomNonbondedForce.addParticle`.

    Parameters
    ----------
    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance :math:`r_\mathrm{cut}` for all nonbonded
        interactions in the simulation sytem. Must be less than half the
        minimum periodic simulation box dimension.

        **Reference unit**: :math:`\mathrm{nm}`.

    powers : `tuple` or `dict`, keyword-only, default: :code:`(12, 6)`
        Powers :math:`\gamma_\mathrm{r}` and :math:`\gamma_\mathrm{a}`,
        in that order. If a `dict` is provided, use the keys :code:`"r"`
        and :code:`"a"`, respectively.

    mix : `str`, keyword-only, default: :code:`"arithmetic"`
        Mixing rule for :math:`\sigma_{12}` and :math:`\epsilon_{12}`.

        .. container::

           **Valid values**:

           * :code:`"arithmetic"`: Lorentz-Berthelot combining rule.

             .. math::

                \begin{gather*}
                  \sigma_{12}=\frac{\sigma_1+\sigma_2}{2}\\
                  \epsilon_{12}=\sqrt{\epsilon_1\epsilon_2}
                \end{gather*}

             **Per-particle parameters**: :math:`(\sigma_i,\,\epsilon_i)`.

           * :code:`"geometric"`: Geometric mixing rule.

             .. math::

                \begin{gather*}
                  \sigma_{12}=\sqrt{\sigma_1\sigma_2}\\
                  \epsilon_{12}=\sqrt{\epsilon_1\epsilon_2}
                \end{gather*}

             **Per-particle parameters**: :math:`(\sigma_i,\,\epsilon_i)`.

           * :code:`"sixthpower"`: Sixth-power mixing rule.

             .. math::

                \begin{gather*}
                  \sigma_{12}=\left(\frac{\sigma_1^6
                  +\sigma_2^6}{2}\right)^{1/6}\\
                  \epsilon_{12}=\frac{2\sqrt{\epsilon_1\epsilon_2}
                  \sigma_1^3\sigma_2^3}{\sigma_1^6+\sigma_2^6}
                \end{gather*}

             **Per-particle parameters**: :math:`(\sigma_i,\,\epsilon_i)`.

           * :code:`"sigma12 = ...; epsilon12 = ...;"`: Custom mixing
             rule. The string containing the expression for
             :math:`\sigma_{12}` and :math:`\epsilon_{12}` must be
             written in valid C++ syntax, with any custom global and
             per-particle parameters and tabulated functions defined in
             `global_params`, `per_params`, and `tab_funcs`,
             respectively.

    global_params : `dict`, keyword-only, optional
        Additional global parameters for use in the definition of
        :math:`\sigma_{12}` and :math:`\epsilon_{12}`.

    per_params : `list`, keyword-only, optional
        Additional per-particle parameters for use in the definition of
        :math:`\sigma_{12}` and :math:`\epsilon_{12}`.

    tab_funcs : `dict`, keyword-only, optional
        Optional tabulated functions for use in the definition of
        :math:`\sigma_{12}` and :math:`\epsilon_{12}`.

    Returns
    -------
    pair_wca : `openmm.CustomNonbondedForce`
        WCA pair potential.
    """

    return ljts(
        cutoff,
        powers=powers,
        mix=mix,
        wca=True,
        global_params=global_params,
        per_params=per_params,
        tab_funcs=tab_funcs,
    )


def yukawa(
    cutoff: Union[float, unit.Quantity],
    cutoff_yukawa: Union[float, unit.Quantity] = None,
    *,
    shift: bool = True,
    mix: str = "geometric",
    per_params: list = None,
    global_params: dict[str, Union[float, unit.Quantity]] = None,
    tab_funcs: dict[
        str,
        Union[
            np.ndarray[Union[int, float]],
            openmm.unit.Quantity,
            openmm.Discrete2DFunction,
        ],
    ] = None,
) -> openmm.CustomNonbondedForce:
    r"""
    Implements the generalized intermolecular Yukawa pair potential.

    The potential energy between two Yukawa particles is given by

    .. math::

       u_\mathrm{Yukawa}(r_{12})=\frac{\alpha_{12}
       \exp(-\kappa r_{12})}{r_{12}}

    where :math:`\alpha_{12}` is a magnitude scaling constant in
    :math:`\mathrm{nm}\cdot\mathrm{kJ/mol}` and :math:`\kappa` is a
    screening length in :math:`\mathrm{nm}^{-1}`.

    After creating the pair potential, particles should be registered
    using :meth:`openmm.openmm.CustomNonbondedForce.addParticle`.

    Parameters
    ----------
    cutoff : `float` or `openmm.unit.Quantity`
        Shared cutoff distance for all nonbonded interactions in the
        simulation sytem. Must be less than half the minimum periodic
        simulation box dimension.

        **Reference unit**: :math:`\mathrm{nm}`.

    cutoff_yukawa : `float` or `openmm.unit.Quantity`, optional
        Cutoff distance :math:`r_\mathrm{cut}` for the Yukawa potential.
        Must be less than the shared cutoff distance. If not provided,
        it is set to the shared cutoff distance.

        **Reference unit**: :math:`\mathrm{nm}`.

    shift : `bool`, keyword-only, default: :code:`True`
        Determines whether the Yukawa potential is shifted at its
        cutoff to :math:`0\,\mathrm{kJ/mol}`.

    mix : `str`, keyword-only, default: :code:`"geometric"`
        Mixing rule for :math:`\alpha_{12}`.

        .. container::

           **Valid values**:

           * :code:`"geometric"`: Geometric mixing rule.

             .. math::

                \alpha_{12}=\sqrt{\alpha_1\alpha_2}

             **Global parameter**: :math:`\kappa`.

             **Per-particle parameter**: :math:`\alpha_i`.

           * :code:`"alpha12 = ...;"`: Custom mixing rule. The string
             containing the expressions for :math:`\alpha_{12}` (and
             optionally, :math:`\kappa`) must be written in valid C++
             syntax, with any custom global and per-particle parameters
             and tabulated functions defined in `global_params`,
             `per_params`, and `tab_funcs`, respectively.

             Should you want to use the geometric mixing rule for
             :math:`\alpha_{12}` and specify an alternate definition for
             :math:`\kappa`, you can include :code:`"geometric;"` in the
             string but

             Be aware that the
             parameters and functions used in the geometric mixing rule
             will take precedence over those defined here.

    global_params : `dict`, keyword-only, optional
        Additional global parameters for use in the definition of
        :math:`\alpha_{12}` (and optionally, :math:`\kappa`).

    per_params : `list`, keyword-only, optional
        Additional per-particle parameters for use in the definition of
        :math:`\alpha_{12}` (and optionally, :math:`\kappa`).

    tab_funcs : `dict`, keyword-only, optional
        Optional tabulated functions for use in the definition of
        :math:`\alpha_{12}` (and optionally, :math:`\kappa`).

    Returns
    -------
    pair_yukawa : `openmm.CustomNonbondedForce`
        Yukawa pair potential.
    """

    if isinstance(cutoff, unit.Quantity):
        cutoff = cutoff.value_in_unit(unit.nanometer)
    if cutoff_yukawa is None:
        cutoff_yukawa = cutoff
    else:
        if isinstance(cutoff_yukawa, unit.Quantity):
            cutoff_yukawa = cutoff_yukawa.value_in_unit(unit.nanometer)
        if cutoff_yukawa > cutoff:
            emsg = (
                "The cutoff distance for the LJTS potential must be "
                "less than the shared cutoff distance."
            )
            raise ValueError(emsg)

    prefix = f"step({cutoff_yukawa}-r)*(" if cutoff != cutoff_yukawa else "("
    root = "alpha12*exp(-kappa*r)/r"
    suffix = (
        f"-ucut);ucut=alpha12*exp(-kappa*{cutoff_yukawa})/{cutoff_yukawa};"
        if shift
        else ");"
    )
    if "geometric" in mix:
        mix = mix.replace("geometric", "alpha12=sqrt(alpha1*alpha2)")
        if not mix.endswith(";"):
            mix += ";"
        if "kappa" not in mix and "kappa" not in global_params:
            raise ValueError("Global parameter 'kappa' not defined.")
        if per_params is None:
            per_params = []
        per_params += ["alpha"]

    pair_yukawa = openmm.CustomNonbondedForce(f"{prefix}{root}{suffix}{mix}")
    _setup_pair(pair_yukawa, cutoff, global_params, per_params, tab_funcs)
    return pair_yukawa



import os
import pathlib
import sys
import tarfile
import urllib
import warnings

import ase.io
import dynasor
import MDAnalysis as mda
from MDAnalysis.analysis.rdf import InterRDF
from MDAnalysis.tests.datafiles import TPR, XTC
import numpy as np

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.analysis import structure # noqa: E402

warnings.filterwarnings("ignore")
rng = np.random.default_rng()

def test_func_radial_histogram():

    L = 20
    half_L = L // 2
    dimensions = np.array((L, L, L, 90, 90, 90), dtype=int)
    origin = half_L * np.ones(3)

    N = 1_000
    norm = L // 2 * rng.random(N)
    counts = np.histogram(norm, bins=half_L, range=(0, half_L + 1))[0]
    neighbors = rng.random((N, 3))
    neighbors *= norm[:, None] / np.linalg.norm(neighbors, axis=1,
                                                keepdims=True)
    neighbors += dimensions[:3] / 2

    # TEST CASE 1: Radial histogram for randomly placed particles
    assert np.array_equal(
        counts,
        structure.radial_histogram(
            origin,
            neighbors,
            n_bins=half_L,
            range_=(0, half_L + 1),
            dimensions=dimensions
        )
    )

def test_func_radial_fourier_transform():

    alpha = 1 + 9 * rng.random()
    r = np.linspace(1e-8, 20, 2_000)
    q = 1 / r
    f = np.exp(-alpha * r) / r
    F = 4 * np.pi / (alpha ** 2 + q ** 2)

    # TEST CASE 1: Radial Fourier transform of function exp(-ar)/r,
    # which has analytical form 4*pi/(a^2+q^2)
    assert np.allclose(F, structure.radial_fourier_transform(r, f, q),
                       atol=4e-5)

def test_class_rdf():

    """
    The following test cases are adapted from the "Average radial
    distribution functions" page from the MDAnalysis User Guide
    (https://userguide.mdanalysis.org/stable/examples/analysis/structure/average_rdf.html).
    """

    universe = mda.Universe(TPR, XTC)
    res60 = universe.select_atoms("resid 60")
    water = universe.select_atoms("resname SOL")
    thr = universe.select_atoms("resname THR")
    n_bins = 75

    ### residue60/water
    irdf = InterRDF(res60, water, nbins=n_bins).run()

    # TEST CASE 1: Batched serial RDF calculation
    rdf = structure.RadialDistributionFunction(
        res60, water, n_bins=n_bins, n_batches=2
    ).run()
    assert np.allclose(irdf.results.bins, rdf.results.bins)
    assert np.allclose(irdf.results.rdf, rdf.results.rdf)

    # TEST CASE 2: Batched parallel RDF calculation
    prdf = structure.RadialDistributionFunction(
        res60, water, n_bins=n_bins, n_batches=2, parallel=True
    ).run(module="joblib", n_jobs=1)
    assert np.allclose(irdf.results.bins, prdf.results.bins)
    assert np.allclose(irdf.results.rdf, prdf.results.rdf)

    ### residue60/residue60 w/ self exclusion
    exclusion = (1, 1)
    irdf = InterRDF(res60, res60, nbins=n_bins, exclusion_block=exclusion).run()

    # TEST CASE 1: Serial RDF calculation
    rdf = structure.RadialDistributionFunction(
        res60, n_bins=n_bins, exclusion=exclusion
    ).run()
    assert np.allclose(irdf.results.bins, rdf.results.bins)
    assert np.allclose(irdf.results.rdf, rdf.results.rdf)

    # TEST CASE 2: Parallel RDF calculation
    prdf = structure.RadialDistributionFunction(
        res60, n_bins=n_bins, exclusion=exclusion, parallel=True
    ).run(module="joblib", n_jobs=1)
    assert np.allclose(irdf.results.bins, prdf.results.bins)
    assert np.allclose(irdf.results.rdf, prdf.results.rdf)

    ### threonine/threonine w/ self exclusion
    exclusion = (14, 14)
    irdf = InterRDF(thr, thr, nbins=n_bins, exclusion_block=exclusion).run()

    # TEST CASE 1: Serial RDF calculation
    rdf = structure.RadialDistributionFunction(
        thr, n_bins=n_bins, exclusion=exclusion
    ).run()
    assert np.allclose(irdf.results.bins, rdf.results.bins)
    assert np.allclose(irdf.results.rdf, rdf.results.rdf)

    # TEST CASE 2: Parallel RDF calculation
    prdf = structure.RadialDistributionFunction(
        thr, n_bins=n_bins, exclusion=exclusion, parallel=True
    ).run(module="joblib", n_jobs=1)
    assert np.allclose(irdf.results.bins, prdf.results.bins)
    assert np.allclose(irdf.results.rdf, prdf.results.rdf)

    ### threonine/threonine w/ carbon exclusion
    exclusion = (4, 10)
    irdf = InterRDF(thr, thr, nbins=n_bins, exclusion_block=exclusion).run()

    # TEST CASE 1: Serial RDF calculation
    rdf = structure.RadialDistributionFunction(
        thr, n_bins=n_bins, exclusion=exclusion
    ).run()
    assert np.allclose(irdf.results.bins, rdf.results.bins)
    assert np.allclose(irdf.results.rdf, rdf.results.rdf)

    # TEST CASE 2: Parallel RDF calculation
    prdf = structure.RadialDistributionFunction(
        thr, n_bins=n_bins, exclusion=exclusion, parallel=True
    ).run(module="joblib", n_jobs=1)
    assert np.allclose(irdf.results.bins, prdf.results.bins)
    assert np.allclose(irdf.results.rdf, prdf.results.rdf)

def test_class_structurefactor():

    """
    The following test cases are adapted from the "Static structure
    factor in halide perovskite (CsPbI3)" page from the dynasor documentation
    (https://dynasor.materialsmodeling.org/dev/tutorials/static_structure_factor.html).
    """

    path = os.getcwd()
    if "tests" in path:
        path_split = path.split("/")
        path = "/".join(path_split[:path_split.index("tests") + 1])
    else:
        path += "/tests"
    if not os.path.isdir(f"{path}/data/ssf"):
        os.makedirs(f"{path}/data/ssf")
    os.chdir(f"{path}/data/ssf")

    if not os.path.isdir("md_runs"):
        with urllib.request.urlopen(
                "https://zenodo.org/records/10149723/files/md_runs.tar.gz"
            ) as r:
            with open("md_runs.tar.gz", "wb") as f:
                f.write(r.read())
        with tarfile.open("md_runs.tar.gz", "r:gz") as tar:
            tar.extractall()
        os.remove("md_runs.tar.gz")
    os.chdir("md_runs/NVT_tetra_size8_T450_nframes1000")

    q_max = 2.2
    stop = 10
    atoms = ase.io.read("model.xyz")
    atomic_indices = atoms.symbols.indices()
    atom_types = sorted(atomic_indices.keys())
    traj = dynasor.Trajectory("movie.nc", trajectory_format="nc",
                              atomic_indices=atomic_indices,
                              frame_stop=stop)
    q_points = dynasor.get_spherical_qpoints(traj.cell, q_max=q_max)
    sample = dynasor.compute_static_structure_factors(traj, q_points)
    q_norms = np.linalg.norm(sample.q_points, axis=1)
    q_norms_unique = np.unique(np.round(q_norms, 11))

    universe = mda.Universe("model.xyz", "movie.nc")
    groups = [universe.select_atoms(f"element {e}") for e in atom_types]

    # TEST CASE 1: Partial structure factors
    psf_exp = structure.StructureFactor(
        groups, mode="partial", q_max=q_max, parallel=True
    ).run(stop=stop)
    psf_trig = structure.StructureFactor(
        groups, mode="partial", q_max=q_max, form="trig", parallel=True
    ).run(stop=stop)
    psf_dynasor = np.empty((len(psf_exp.results.pairs), len(q_norms_unique)))
    for i, (j, k) in enumerate(psf_exp.results.pairs):
        psf_dynasor[i] = np.fromiter(
            (sample[f"Sq_{atom_types[j]}_{atom_types[k]}"]
             [np.isclose(q, q_norms)].mean() for q in q_norms_unique),
            dtype=float,
            count=len(q_norms_unique)
        )
    assert np.allclose(psf_exp.results.wavenumbers, q_norms_unique)
    assert np.allclose(psf_exp.results.ssf, psf_dynasor, atol=1e-6)
    assert np.allclose(psf_trig.results.ssf, psf_dynasor, atol=1e-6)

    # TEST CASE 2: Static structure factors
    ssf_exp = structure.StructureFactor(
        groups, wavevectors=q_points, sort=False, unique=False, parallel=True
    ).run(stop=stop)
    ssf_trig = structure.StructureFactor(
        groups, form="trig", wavevectors=q_points, sort=False, unique=False,
        parallel=True
    ).run(stop=stop)
    assert np.allclose(ssf_exp.results.ssf[0], sample.Sq[:, 0])
    assert np.allclose(ssf_trig.results.ssf[0], sample.Sq[:, 0])

def test_class_intermediatescatteringfunction():

    path = os.getcwd()
    if "tests" in path:
        path_split = path.split("/")
        path = "/".join(path_split[:path_split.index("tests") + 1])
    else:
        path += "/tests"
    if not os.path.isdir(f"{path}/data/ssf"):
        os.makedirs(f"{path}/data/ssf")
    os.chdir(f"{path}/data/ssf")

    if not os.path.isdir("md_runs"):
        with urllib.request.urlopen(
                "https://zenodo.org/records/10149723/files/md_runs.tar.gz"
            ) as r:
            with open("md_runs.tar.gz", "wb") as f:
                f.write(r.read())
        with tarfile.open("md_runs.tar.gz", "r:gz") as tar:
            tar.extractall()
        os.remove("md_runs.tar.gz")
    os.chdir("md_runs/NVT_tetra_size8_T450_nframes1000")

    stop = 20
    n_lags = stop // 2
    atoms = ase.io.read("model.xyz")
    atomic_indices = atoms.symbols.indices()
    atom_types = sorted(atomic_indices.keys())
    traj = dynasor.Trajectory("movie.nc", trajectory_format="nc",
                              atomic_indices=atomic_indices,
                              frame_stop=stop)
    q_points = dynasor.get_spherical_qpoints(traj.cell, q_max=2,
                                             max_points=2_000)
    sample = dynasor.compute_dynamic_structure_factors(
        traj, q_points, dt=1, window_size=n_lags, calculate_incoherent=True
    )
    q_norms = np.linalg.norm(sample.q_points, axis=1)
    q_norms_unique = np.unique(np.round(q_norms, 11))

    universe = mda.Universe("model.xyz", "movie.nc")
    groups = [universe.select_atoms(f"element {e}") for e in atom_types]
    isf_exp = structure.IntermediateScatteringFunction(
        groups, mode="partial", wavevectors=q_points, n_lags=n_lags,
        incoherent=True, parallel=True
    ).run(stop=stop)
    isf_trig = structure.IntermediateScatteringFunction(
        groups, mode="partial", form="trig", wavevectors=q_points,
        n_lags=n_lags, incoherent=True, parallel=True
    ).run(stop=stop)

    # TEST CASE 1: Partial coherent intermediate scattering functions
    cisf_dynasor = np.empty((n_lags, len(isf_exp.results.pairs),
                             len(q_norms_unique)))
    for i, (j, k) in enumerate(isf_exp.results.pairs):
        for iq, q in enumerate(q_norms_unique):
            cisf_dynasor[:, i, iq] = (
                sample[f"Fqt_coh_{atom_types[j]}_{atom_types[k]}"]
                [np.isclose(q, q_norms), :n_lags].mean(axis=0)
            )
    assert np.allclose(isf_exp.results.cisf, cisf_dynasor)
    assert np.allclose(isf_trig.results.cisf, cisf_dynasor)

    # TEST CASE 2: Partial incoherent intermediate scattering functions
    iisf_dynasor = np.empty((n_lags, len(atom_types), len(q_norms_unique)))
    for i in range(len(atom_types)):
        for iq, q in enumerate(q_norms_unique):
            iisf_dynasor[:, i, iq] = (
                sample[f"Fqt_incoh_{atom_types[i]}"]
                [np.isclose(q, q_norms), :n_lags].mean(axis=0)
            )
    assert np.allclose(isf_exp.results.iisf, iisf_dynasor)
    assert np.allclose(isf_trig.results.iisf, iisf_dynasor)

    isf_exp = structure.IntermediateScatteringFunction(
        groups, wavevectors=q_points, n_lags=n_lags, incoherent=True,
        sort=False, unique=False, parallel=True
    ).run(stop=stop)
    isf_trig = structure.IntermediateScatteringFunction(
        groups, form="trig", wavevectors=q_points, n_lags=n_lags,
        incoherent=True, sort=False, unique=False, parallel=True
    ).run(stop=stop)

    # TEST CASE 3: Coherent intermediate scattering function
    assert np.allclose(isf_exp.results.cisf[:, 0], sample.Fqt_coh[:, :n_lags].T)
    assert np.allclose(isf_trig.results.cisf[:, 0], sample.Fqt_coh[:, :n_lags].T)

    # TEST CASE 4: Incoherent intermediate scattering function
    assert np.allclose(isf_exp.results.iisf[:, 0], sample.Fqt_incoh[:, :n_lags].T)
    assert np.allclose(isf_trig.results.iisf[:, 0], sample.Fqt_incoh[:, :n_lags].T)


"""
Electrostatic properties
========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains classes to quantify electrostatic properties, such
as the instantaneous dipole moment.
"""

from numbers import Real
from typing import Union

import MDAnalysis as mda
import numpy as np

from .base import DynamicAnalysisBase
from .. import FOUND_OPENMM, Q_, ureg
from ..algorithm.topology import unwrap
from ..algorithm.unit import is_unitless, strip_unit

if FOUND_OPENMM:
    from openmm import unit


def calculate_relative_permittivity(
    M: np.ndarray[float], temperature: float, volume: float, *, reduced: bool = False
) -> float:
    r"""
    Calculates the relative permittivity (or static dielectric constant)
    :math:`\varepsilon_\mathrm{r}` of a medium using dipole moments
    :math:`\mathbf{M}`.

    The dipole moment fluctuation formula [1]_ relates the relative
    permittivity to the dipole moment via

    .. math::

       \varepsilon_\mathrm{r}=1+\frac{\overline{\langle\mathbf{M}^2\rangle
       -\langle\mathbf{M}\rangle^2}}{3\varepsilon_0 Vk_\mathrm{B}T}

    where the angular brackets :math:`\langle\,\cdot\,\rangle` denote
    the ensemble average, the overline signifies the spatial average,
    :math:`\varepsilon_0` is the vacuum permittivity,
    :math:`k_\mathrm{B}` is the Boltzmann constant, and :math:`T` is
    the system temperature.

    .. note::

       If residues (molecules) in your system have net charges, the
       dipole moments must be made position-independent by subtracting
       the product of the net charges and the centers of mass or
       geometry.

    Parameters
    ----------
    M : array-like
        Dipole moment vectors :math:`\mathbf{M}`.

        **Shape**: :math:`(N_\mathrm{frames},\,3)`.

        **Reference unit**: :math:`\mathrm{e\cdotÅ}`.

    temperature : `float`
        System temperature :math:`T`.

        .. note::

           If :code:`reduced=True`, `temperature` should be equal to the
           energy scale. When the Lennard-Jones potential is used, it
           generally means that :math:`T^*=1`, or `temperature=1`.

        **Reference unit**: :math:`\mathrm{K}`.

    volume : `float`
        System volume :math:`V`.

        **Reference unit**: :math:`\mathrm{Å^3}`.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units.

    Returns
    -------
    dielectric : `float`
        Relative permittivity (or static dielectric constant).

    References
    ----------
    .. [1] Neumann, M. Dipole Moment Fluctuation Formulas in Computer
       Simulations of Polar Systems. *Molecular Physics* **1983**,
       *50* (4), 841–858. https://doi.org/10.1080/00268978300102721.
    """

    conversion_factor = (
        4 * np.pi
        if reduced
        else (
            1
            * ureg.elementary_charge**2
            / (
                ureg.vacuum_permittivity
                * ureg.angstrom
                * ureg.boltzmann_constant
                * ureg.kelvin
            )
        ).m_as(ureg.dimensionless)
    )

    return 1 + (
        conversion_factor * (M**2 - M.mean(axis=0) ** 2).mean() / (volume * temperature)
    )


class DipoleMoment(DynamicAnalysisBase):
    """
    Serial and parallel implementations to calculate the dipole moment
    :math:`\\mathbf{M}`.

    For a system with :math:`N` atoms or molecules, the dipole moment is
    given by

    .. math::

       \\mathbf{M}=\\sum_i^{N}q_i\\mathbf{z}_i

    where :math:`q_i` and :math:`\\mathbf{z}_i` are the charge and
    position of entity :math:`i`.

    The dipole moment can be used to estimate the relative permittivity
    (or static dielectric constant) via the dipole moment fluctuation
    formula [1]_:

    .. math::

       \\varepsilon_\\mathrm{r}=1+\\frac{\\overline{
       \\langle\\mathbf{M}^2\\rangle-\\langle\\mathbf{M}\\rangle^2
       }}{3\\varepsilon_0 Vk_\\mathrm{B}T}

    where the angular brackets :math:`\\langle\\,\\cdot\\,\\rangle` denote
    the ensemble average, the overline signifies the time average,
    :math:`\\varepsilon_0` is the vacuum permittivity,
    :math:`k_\\mathrm{B}` is the Boltzmann constant, and :math:`T` is
    the temperature.

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Groups of atoms for which the dipole moments are calculated.

        .. note::

           If :code:`neutralize=True` or :code:`unwrap=True`, all atoms
           of any particular molecule must belong to the same atom
           group.

    charges : array-like, `openmm.unit.Quantity`, or \
    `pint.Quantity`, keyword-only,optional
        Charges :math:`q_i` for the atoms in the
        :math:`N_\\mathrm{groups}` groups in `groups`. If not provided,
        it should be available in and will be retrieved from the main
        :class:`MDAnalysis.core.universe.Universe` object.

        **Shape**: :math:`(N_\\mathrm{groups},)` array of real numbers
        or :math:`(N_i,)` arrays, where :math:`N_i` is the number of
        atoms in group :math:`i`.

        **Reference unit**: :math:`\\mathrm{e}`.

    dim_scales : `float` or array-like, keyword-only, optional
        Scale factors for the system dimensions. If an `int` is
        provided, the same value is used for all axes.

        **Shape**: :math:`(3,)`.

    average : `bool`, keyword-only, default: :code:`False`
        Determines whether the dipole moment vectors and volumes are
        averaged over the :math:`N_\\mathrm{frames}` analysis frames.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units. Only affects
        :meth:`calculate_relative_permittivity` calls.

    neutralize : `bool`, keyword-only, default: :code:`False`
        Specifies whether net charges of molecules are subtracted at the
        center of mass. Must be enabled if your system contains
        molecules with net charges and you want to calculate the
        relative permittivity, but should be disabled if you are only
        interested in the dipole moment.

        .. note::

           The topology must contain residue (molecule) information.

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the simulation system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for `results.times`, call
        :code:`results.units["times"]`.

    results.times : `numpy.ndarray`
        Times :math:`t`. Only available if :code:`average=False`.

        **Shape**: :math:`(N_\\mathrm{frames},)`.

        **Reference unit**: :math:`\\mathrm{ps}`.

    results.dipoles : `numpy.ndarray`
        Dipole moment vectors :math:`\\mathbf{M}`.

        **Shape**: :math:`(N_\\mathrm{groups},\\,3)` or
        :math:`(N_\\mathrm{frames},\\,N_\\mathrm{groups},\\,3)`.

        **Reference unit**: :math:`\\mathrm{e\\cdotÅ}`.

    results.volumes : `numpy.ndarray`
        System volumes :math:`V`.

        **Shape**: :math:`(N_\\mathrm{frames},)`.

        **Reference unit**: :math:`\\mathrm{Å^3}`.

    results.dielectrics : `numpy.ndarray`
        Relative permittivity (or static dielectric constant)
        :math:`\\varepsilon_\\mathrm{r}` in each dimension.

        **Shape**: :math:`(3,)`.
        
    Example
    --------
    First, this analysis class must be imported:

    >>> from mdcraft.analysis.electrostatics import DipoleMoment

    Then, after loading a simulation trajectory:

    >>> universe = mda.Universe("simulation.nc", "topology.cif")

    We must then select the atom-groups to be analyzed (typically the solvent):

    >>> ag = universe.select_atoms("resname SOL")
    
    The `DipoleMoment` class can be instantiated with the selected atom-groups, remembering to specify the charges:
    
    >>> dpm = DipoleMoment(ag, charges=ag.charges)
    >>> dpm.run()

    The results can be obtained under the `results` attribute:
    
    >>> dpm.results.dipoles
    
    To calculate the relative permittivity, the temperature must be provided:
    
    >>> dpm.calculate_relative_permittivity(300)
    
    These results can be saved to a file using the `save` method:
        
    >>> dpm.save("dpm")

    References
    ----------
    .. [1] Neumann, M. Dipole Moment Fluctuation Formulas in Computer
       Simulations of Polar Systems. *Molecular Physics* **1983**,
       *50* (4), 841–858. https://doi.org/10.1080/00268978300102721.
    """

    def __init__(
        self,
        groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
        charges: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
        dim_scales: Union[float, tuple[float]] = 1,
        average: bool = False,
        reduced: bool = False,
        neutralize: bool = False,
        unwrap: bool = False,
        parallel: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> None:

        self._groups = [groups] if isinstance(groups, mda.AtomGroup) else groups
        self._n_groups = len(self._groups)
        self.universe = self._groups[0].universe
        if self.universe.dimensions is None:
            raise ValueError(
                "Trajectory does not contain system " "dimension information."
            )
        super().__init__(self.universe.trajectory, parallel, verbose, **kwargs)

        if isinstance(dim_scales, Real) or (
            len(dim_scales) == 3 and all(isinstance(f, Real) for f in dim_scales)
        ):
            self._dim_scales = dim_scales
        else:
            emsg = (
                "'dim_scales' must be a floating-point number "
                "or an array with shape (3,)."
            )
            raise ValueError(emsg)

        self._Ns = np.fromiter(
            (ag.n_atoms for ag in self._groups), dtype=int, count=self._n_groups
        )
        self._N = self._Ns.sum()
        self._slices = []
        _ = 0
        for N in self._Ns:
            self._slices.append(slice(_, _ + N))
            _ += N

        if charges is not None:
            charges = list(charges)
            if len(charges) == self._n_groups:
                for i, (ag, q) in enumerate(zip(self._groups, charges)):
                    charges[i] = strip_unit(q, "e")[0]
                    if isinstance(charges[i], Real):
                        charges[i] *= np.ones(ag.n_atoms)
                    elif ag.n_atoms != len(charges[i]):
                        emsg = (
                            "The number of charges in "
                            f"'charges[{i}]' is not equal to the "
                            "number of atoms in the corresponding "
                            "group."
                        )
                        raise ValueError(emsg)
                self._charges = charges
            else:
                emsg = (
                    "The number of group charge arrays is not "
                    "equal to the number of groups."
                )
                raise ValueError(emsg)
        elif hasattr(self.universe.atoms, "charges"):
            self._charges = [ag.charges for ag in self._groups]
        else:
            raise ValueError("The topology has no charge information.")

        self._all_neutral = np.allclose(
            self.universe.atoms.total_charge(compound="residues"), 0, atol=1e-6
        )

        self._average = average
        self._neutralize = neutralize
        self._reduced = reduced
        self._unwrap = unwrap
        self._verbose = verbose

    def _prepare(self) -> None:

        if self._unwrap:

            # Navigate to first frame in analysis
            ts = self._sliced_trajectory[0]

            # Get (scaled) system dimensions
            dimensions = self._dim_scales * ts.dimensions[:3]

            # Preallocate arrays to determine the number of periodic
            # boundary crossings for each atom
            self._positions_old = np.empty((self._N, 3))
            for ag, s in zip(self._groups, self._slices):

                # Locally unwrap molecules at the edge of the simulation
                # system
                self._positions_old[s] = ag.unwrap()

            self._images = np.zeros((self._N, 3), dtype=int)

            # Store unwrapped particle positions in a shared memory array
            # for parallel analysis
            if self._parallel:
                self._positions = np.empty((self.n_frames, self._N, 3))
                for i, _ in enumerate(self._sliced_trajectory):
                    for ag, s in zip(self._groups, self._slices):
                        self._positions[i, s] = ag.positions

                    # Globally unwrap atom positions
                    unwrap(
                        self._positions[i],
                        self._positions_old,
                        dimensions,
                        thresholds=dimensions / 2,
                        images=self._images,
                    )

        # Store reference units
        self.results.units = {
            "dipoles": ureg.elementary_charge * ureg.angstrom,
            "volumes": ureg.angstrom**3,
        }

        # Preallocate arrays to store dipole moments and system volumes
        self.results.dipoles = np.empty((self.n_frames, self._n_groups, 3))
        self.results.volumes = np.empty(self.n_frames)

        # Store time information, if necessary
        if not self._average:
            self.results.times = np.fromiter(
                (ts.time for ts in self._sliced_trajectory),
                dtype=float,
                count=self.n_frames,
            )
            self.results.units["times"] = ureg.picosecond

        # Preallocate array to hold atom positions for a given frame
        # (so that one doesn't have to be recreated each frame)
        if not self._parallel:
            self._positions = np.empty((self._N, 3))

    def _single_frame(self) -> None:

        # Get (scaled) system dimensions
        if self._unwrap:
            dimensions = self._dim_scales * self._ts.dimensions[:3]

        for i, (ag, s, q) in enumerate(zip(self._groups, self._slices, self._charges)):

            # Store atom positions in the current frame
            self._positions[s] = ag.positions

            # Globally unwrap atom positions
            if self._unwrap:
                unwrap(
                    self._positions[s],
                    self._positions_old[s],
                    dimensions,
                    thresholds=dimensions / 2,
                    images=self._images[s],
                )

            # Subtract the net charge at the center of mass of each
            # molecule, if necessary
            if self._neutralize:
                q -= q * np.concatenate([r.atoms.masses / r.mass for r in ag.residues])

            # Calculate the dipole moment for the current frame
            self.results.dipoles[self._frame_index, i] = q @ self._positions[s]

        # Store the system volume for the current frame
        self.results.volumes[self._frame_index] = (
            np.prod(self._dim_scales) * self._ts.volume
        )

    def _single_frame_parallel(
        self, index: int
    ) -> tuple[int, float, np.ndarray[float]]:

        # Set current trajectory frame
        ts = self._sliced_trajectory[index]

        # Preallocate arrays to hold
        dipoles = np.empty((self._n_groups, 3))

        for i, (ag, s, q) in enumerate(zip(self._groups, self._slices, self._charges)):

            # Get atom positions in the current frame
            positions = self._positions[index, s] if self._unwrap else ag.positions

            # Subtract the net charge at the center of mass of each
            # molecule, if necessary
            if self._neutralize:
                q -= q * np.concatenate([r.atoms.masses / r.mass for r in ag.residues])

            # Calculate the dipole moment for the current frame
            dipoles[i] = q @ positions

        return index, np.prod(self._dim_scales) * ts.volume, dipoles

    def _conclude(self) -> None:

        # Consolidate parallel results and clean up memory by deleting
        # arrays that will not be reused
        if self._parallel:
            self._results = sorted(self._results)
            self.results.volumes = np.fromiter(
                (r[1] for r in self._results), dtype=float, count=self.n_frames
            )
            self.results.dipoles = np.stack([r[2] for r in self._results])

            del self._results
            if self._unwrap:
                del self._positions
        else:
            del self._positions
        if self._unwrap:
            del self._positions_old, self._images

        # Average results, if requested
        if self._average:
            self.results.dipoles = self.results.dipoles.mean(axis=0)
            self.results.volumes = self.results.volumes.mean()

    def calculate_relative_permittivity(
        self, temperature: Union[float, "unit.Quantity", Q_]
    ) -> None:
        r"""
        Calculates the relative permittivity (or static dielectric
        constant) :math:`\varepsilon_\mathrm{r}` of a medium using
        dipole moments :math:`\mathbf{M}`.

        Parameters
        ----------
        temperature : `float`, `openmm.unit.Quantity`, or `pint.Quantity`
            System temperature :math:`T`.

            .. note::

               If :code:`reduced=True` was set in the
               :class:`DipoleMoment` constructor, `temperature` should
               be equal to the energy scale. When the Lennard-Jones
               potential is used, it generally means that
               :math:`T^*=1`, or `temperature=1`.

            **Reference unit**: :math:`\mathrm{K}`.
        """

        if self._average:
            emsg = (
                "Cannot compute relative permittivity using the"
                "averaged dipole moment."
            )
            raise RuntimeError(emsg)
        elif not self._all_neutral and not self._neutralize:
            emsg = (
                "Cannot compute relative permittivity for a "
                "non-neutral system or a system with ions unless "
                "the net charge is subtracted at the center of "
                "mass of each molecule carrying a net charge."
            )
            raise RuntimeError(emsg)

        if self._reduced and not is_unitless(temperature):
            emsg = "'temperature' cannot have units when reduced=True."
            raise ValueError(emsg)
        temperature = strip_unit(temperature, "K")[0]

        dipoles = self.results.dipoles
        if self._n_groups > 1:
            dipoles = dipoles.sum(axis=dipoles.ndim - 2)
        self.results.dielectrics = calculate_relative_permittivity(
            dipoles, temperature, self.results.volumes.mean(), reduced=self._reduced
        )



"""
Plot colors
===========
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides functions to select colors for plots.
"""

import colorsys
from typing import Union

import matplotlib.colors as mc


def adjust_lightness(
    colors: Union[str, tuple[float], list[Union[str, tuple[float]]]], amount: float
) -> Union[tuple[float], list[tuple[float]]]:
    """
    Adjusts the lightness of colors.

    Parameters
    ----------
    color : `str`, `tuple`, or `list`.
        The colors to adjust. A single color can be specified as a tuple
        of normalized RGB values or a string containing its name or a
        hexadecimal value. Multiple colors can be provided in a `list`.

        **Examples**: :code:`"aquamarine"`, :code:`"#080085"`,
        :code:`(0.269, 0.269, 0.269)`.

    amount : `float`
        The amount to adjust the luminosity by. A value betwen :math:`0`
        and :math:`1` darkens the color, while a value greater than
        :math:`1` lightens the color.

    Returns
    -------
    colors : `tuple` or `list`
        The adjusted colors.
    """

    if isinstance(colors, list):
        for i, color in enumerate(colors):
            colors[i] = adjust_lightness(color, amount)
        return colors

    colors = colorsys.rgb_to_hls(
        *mc.to_rgb(mc.cnames[colors] if colors in mc.cnames else colors)
    )
    return colorsys.hls_to_rgb(colors[0], max(0, min(1, amount * colors[1])), colors[2])



import pathlib
import sys

import numpy as np
import openmm
from openmm import app
import pytest

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.openmm import system as s, topology as t # noqa: E402

def test_func_subset_errors():

    topology = app.Topology()
    topology.addAtom("", "", topology.addResidue("", topology.addChain()))
    positions = np.array((0, 0, 0))

    # TEST CASE 1: Both items to delete and keep are specified
    with pytest.raises(ValueError):
        t.get_subset(topology, positions, delete=[0], keep=[0])

    # TEST CASE 2: No item type specified
    with pytest.raises(ValueError):
        t.get_subset(topology, positions, delete=[0])

def test_func_subset_polymer():

    M = 100
    N_p = 25
    N = M * N_p
    dims = np.array((10, 10, 10))
    positions = t.create_atoms(dims, M * N_p, N_p)
    system = openmm.System()
    topology = app.Topology()
    for _ in range(M):
        chain = topology.addChain()
        s.register_particles(system, topology, N=N_p, chain=chain)
    atoms = list(topology.atoms())
    for m in range(M):
        for n in range(N_p - 1):
            i = m * N_p + n
            topology.addBond(atoms[i], atoms[i + 1])

    # TEST CASE 1: Nothing to do
    topo_sub, pos_sub = t.get_subset(topology, positions)
    assert topo_sub is topology
    assert pos_sub is positions

    n_atoms = 100
    n_chains = n_atoms // N_p
    n_bonds = n_atoms - n_chains

    # TEST CASE 2: Delete everything but the first 100 atoms, with no type specified
    topo_sub, pos_sub = t.get_subset(topology, positions, delete=atoms[n_atoms:])
    assert topo_sub.getNumAtoms() == n_atoms
    assert topo_sub.getNumBonds() == n_bonds
    assert topo_sub.getNumResidues() == n_atoms
    assert topo_sub.getNumChains() == n_chains

    # TEST CASE 3: Delete everything but the first 100 atoms, with types specified
    topo_sub, pos_sub = t.get_subset(topology, positions, delete=np.arange(100, N),
                                 types="atom")
    assert topo_sub.getNumAtoms() == n_atoms
    assert topo_sub.getNumBonds() == n_bonds
    assert topo_sub.getNumResidues() == n_atoms
    assert topo_sub.getNumChains() == n_chains

    # TEST CASE 4: Keep first 100 atoms
    topo_sub, pos_sub = t.get_subset(topology, positions, keep=np.arange(n_atoms),
                                 types="atom")
    assert topo_sub.getNumAtoms() == n_atoms
    assert topo_sub.getNumBonds() == n_bonds
    assert topo_sub.getNumResidues() == n_atoms
    assert topo_sub.getNumChains() == n_chains

    # TEST CASE 5: Keep first 96 bonds
    topo_sub, pos_sub = t.get_subset(topology, positions, keep=np.arange(n_bonds),
                                 types=n_bonds * ["bond"])
    assert topo_sub.getNumAtoms() == n_atoms
    assert topo_sub.getNumBonds() == n_bonds
    assert topo_sub.getNumResidues() == n_atoms
    assert topo_sub.getNumChains() == n_chains

    # TEST CASE 6: Keep first 100 residues
    topo_sub, pos_sub = t.get_subset(topology, positions, keep=np.arange(n_atoms),
                                 types="residue")
    assert topo_sub.getNumAtoms() == n_atoms
    assert topo_sub.getNumBonds() == n_bonds
    assert topo_sub.getNumResidues() == n_atoms
    assert topo_sub.getNumChains() == n_chains

    # TEST CASE 7: Keep first 4 chains
    topo_sub, pos_sub = t.get_subset(topology, positions, keep=np.arange(n_chains),
                                 types="chain")
    assert topo_sub.getNumAtoms() == n_atoms
    assert topo_sub.getNumBonds() == n_bonds
    assert topo_sub.getNumResidues() == n_atoms
    assert topo_sub.getNumChains() == n_chains


"""
Custom OpenMM reporters
=======================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides custom optimized OpenMM reporters.
"""

from typing import Union

import numpy as np
import openmm
from openmm import app, unit

from .file import NetCDFFile


class NetCDFReporter:
    """
    A NetCDF trajectory reporter for OpenMM that can report velocities
    and forces in addition to time and coordinates for all particles in
    the simulation or just a subset.

    Parameters
    ----------
    file : `str`
        Filename of NetCDF file to which the data is saved. If `file`
        does not have the :code:`.nc` extension, it will automatically
        be appended.

    interval : `int`
        Interval (in timesteps) at which to write frames.

    append : `bool`, keyword-only, default: :code:`False`
        If :code:`True`, the existing NetCDF file is opened for data to
        be appended to. If :code:`False`, a new NetCDF file is opened
        (and will clobber an existing file with the same name).

    periodic : `bool`, keyword-only, optional
        Specifies whether particle positions should be translated so the
        center of every molecule lies in the same periodic box. If
        :code:`None` (the default), it will automatically decide whether
        to translate molecules based on whether the system being
        simulated uses periodic boundary conditions.

    velocities : `bool`, keyword-only, default: :code:`False`
        Specifies whether particle velocities should be written to file.

    forces : `bool`, keyword-only, default: :code:`False`
        Specifies whether forces exerted on particles should be written
        to file.

    subset : `slice`, `numpy.ndarray`, or `openmm.app.Topology`, \
    keyword-only, optional
        Slice or array containing the indices of particles to report
        data for. If an OpenMM topology is provided instead, the indices
        are determined from the atoms found in the topology.
    """

    def __init__(
        self,
        file: str,
        interval: int,
        append: bool = False,
        periodic: bool = None,
        *,
        velocities: bool = False,
        forces: bool = False,
        subset: Union[slice, np.ndarray[int], app.Topology] = None,
    ) -> None:

        self._out = NetCDFFile(file, "a" if append else "w")
        self._interval = interval
        self._periodic = periodic
        self._subset = (
            np.fromiter((a.index for a in subset.atoms()), dtype=int)
            if isinstance(subset, app.Topology)
            else subset
        )
        self._velocities = velocities
        self._forces = forces

    def __del__(self) -> None:
        self._out._nc.close()

    def describeNextReport(
        self, simulation: app.Simulation
    ) -> tuple[int, bool, bool, bool, bool, bool]:
        """
        Get information about the next report this NetCDF reporter will
        generate.

        Parameters
        ----------
        simulation : `openmm.app.Simulation`
            OpenMM simulation to generate a report for.

        Returns
        -------
        report : `tuple`
            .. container::

               A six-element tuple containing

               1. the number of steps until the next report,
               2. whether the report will require coordinates,
               3. whether the report will require velocities,
               4. whether the report will require forces,
               5. whether the report will require energies, and
               6. whether coordinates should be wrapped to lie in a
                  single periodic box.
        """

        return (
            self._interval - simulation.currentStep % self._interval,
            True,
            self._velocities,
            self._forces,
            False,
            self._periodic,
        )

    def report(self, simulation: app.Simulation, state: openmm.State) -> None:
        """
        Generate a report.

        Parameters
        ----------
        simulation : `openmm.app.Simulation`
            OpenMM simulation to generate a report for.

        state : `openmm.State`
            Current OpenMM simulation state.
        """

        # Get all requested state data from OpenMM State
        data = {}
        if self._subset is None:
            data["coordinates"] = state.getPositions(asNumpy=True).value_in_unit(
                unit.angstrom
            )
            if self._velocities:
                data["velocities"] = state.getVelocities(asNumpy=True).value_in_unit(
                    unit.angstrom / unit.picosecond
                )
            if self._forces:
                data["forces"] = state.getForces(asNumpy=True).value_in_unit(
                    unit.kilocalorie_per_mole / unit.angstrom
                )
        else:
            data["coordinates"] = state.getPositions(asNumpy=True)[
                self._subset
            ].value_in_unit(unit.angstrom)
            if self._velocities:
                data["velocities"] = state.getVelocities(asNumpy=True)[
                    self._subset
                ].value_in_unit(unit.angstrom / unit.picosecond)
            if self._forces:
                data["forces"] = state.getForces(asNumpy=True)[
                    self._subset
                ].value_in_unit(unit.kilocalorie_per_mole / unit.angstrom)

        # Initialize NetCDF file headers, if not done already
        if not hasattr(self._out._nc, "Conventions"):
            self._out.write_header(
                (
                    simulation.topology.getNumAtoms()
                    if self._subset is None
                    else len(self._subset)
                ),
                simulation.topology.getPeriodicBoxVectors() is not None,
                self._velocities,
                self._forces,
            )

        # Get the lengths and angles that define the size and shape of the
        # simulation box
        pbv = state.getPeriodicBoxVectors()
        if pbv is not None:
            (a, b, c, alpha, beta, gamma) = (
                app.internal.unitcell.computeLengthsAndAngles(pbv)
            )
            data["cell_lengths"] = 10 * np.array((a, b, c))
            data["cell_angles"] = 180 * np.array((alpha, beta, gamma)) / np.pi

        # Write current frame
        self._out.write_model(state.getTime().value_in_unit(unit.picosecond), **data)



import os
import pathlib
import sys

import netCDF4 as nc
import numpy as np
import openmm
from openmm import app, unit
import pytest

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.openmm import file, pair, reporter, system as s, unit as u # noqa: E402

def test_classes_netcdffile_netcdfreporter():

    path = os.getcwd()
    if "tests" in path:
        path_split = path.split("/")
        path = "/".join(path_split[:path_split.index("tests") + 1])
    else:
        path += "/tests"
    if not os.path.isdir(f"{path}/data/netcdf"):
        os.makedirs(f"{path}/data/netcdf")
    os.chdir(f"{path}/data/netcdf")

    # Set up a basic OpenMM simulation for a single LJ particle
    temp = 300 * unit.kelvin
    size = 3.4 * unit.angstrom
    mass = 39.948 * unit.amu
    scales = u.get_lj_scale_factors({
        "energy": (unit.BOLTZMANN_CONSTANT_kB * temp).in_units_of(unit.kilojoule),
        "length": size,
        "mass": mass
    })

    dims = 10 * size * np.ones(3)
    dims_nd = [L / unit.nanometer for L in dims]
    system = openmm.System()
    system.setDefaultPeriodicBoxVectors(
        (dims_nd[0], 0, 0) * unit.nanometer,
        (0, dims_nd[1], 0) * unit.nanometer,
        (0, 0, dims_nd[2]) * unit.nanometer
    )
    topology = app.Topology()
    topology.setUnitCellDimensions(dims)
    pair_lj = pair.lj_coul(dims[0] / 4)
    s.register_particles(system, topology, 1, mass, nbforce=pair_lj, sigma=size,
                         epsilon=21.285 * unit.kilojoule_per_mole)
    system.addForce(pair_lj)

    plat = openmm.Platform.getPlatformByName("CPU")
    dt = 0.005 * scales["time"]
    integrator = openmm.LangevinMiddleIntegrator(temp, 1e-3 / dt, dt)
    simulation = app.Simulation(topology, system, integrator, plat)
    simulation.context.setPositions(dims[None, :] / 2)

    # TEST CASE 1: Correct headers and data for restart file
    # (static method, filename)
    state = simulation.context.getState(getPositions=True, getVelocities=True,
                                        getForces=True)

    file.NetCDFFile.write_file("restart", state)
    ncdf = file.NetCDFFile("restart", "r")
    assert ncdf._nc.Conventions in ("AMBERRESTART", b"AMBERRESTART")
    assert ncdf.get_num_frames() == 1
    assert np.allclose(ncdf.get_positions(), dims / 2)

    # TEST CASE 2: Not a restart file
    ncdf = file.NetCDFFile.write_header("restart.nc", 1, True, True, True)
    with pytest.raises(ValueError):
        ncdf.write_file(state)

    # TEST CASE 3: Correct headers and data for restart file (instance method)
    ncdf = file.NetCDFFile("restart.nc", "w", restart=True)
    ncdf.write_file(state)
    assert ncdf._nc.Conventions in ("AMBERRESTART", b"AMBERRESTART")
    assert ncdf.get_num_frames() == 1
    assert np.allclose(ncdf.get_positions(), dims / 2)
    del ncdf

    # TEST CASE 4: Correct headers and data for restart file
    # (static method, NetCDF file)
    file.NetCDFFile.write_file(
        nc.Dataset("restart.nc", "w", format="NETCDF3_64BIT_OFFSET"),
        state
    )
    ncdf = file.NetCDFFile("restart.nc", "r")
    assert ncdf._nc.Conventions in ("AMBERRESTART", b"AMBERRESTART")
    assert ncdf.get_num_frames() == 1
    assert np.allclose(ncdf.get_positions(), dims / 2)
    del ncdf

    # TEST CASE 5: Correct headers and data for trajectory file
    timesteps = 5
    simulation.reporters.append(
        reporter.NetCDFReporter("traj.nc", 1, periodic=True, velocities=True,
                                forces=True)
    )
    simulation.step(timesteps)

    ncdf = file.NetCDFFile("traj.nc", "r")
    cell_lengths, cell_angles = ncdf.get_dimensions(0)
    assert ncdf._nc.program in ("MDCraft", b"MDCraft")
    assert np.allclose(cell_lengths, dims)
    assert np.allclose(cell_angles, 90 * np.ones(3))
    assert np.allclose(
        ncdf.get_positions(0) - ncdf.get_times(0) * ncdf.get_velocities(0),
        dims / 2,
        atol=2e-3
    )
    assert ncdf.get_num_frames() == timesteps
    assert ncdf.get_velocities().shape == (timesteps, 1, 3)
    assert ncdf.get_forces().shape == (timesteps, 1, 3)

    # TEST CASE 6: Correct number of atoms for subset trajectory file
    s.register_particles(system, topology, 1, mass, nbforce=pair_lj,
                         sigma=size, epsilon=21.285 * unit.kilojoule_per_mole)
    integrator = openmm.LangevinMiddleIntegrator(temp, 1e-3 / dt, dt)
    simulation = app.Simulation(topology, system, integrator, plat)
    simulation.context.setPositions(np.vstack((dims / 4, 3 * dims / 4))
                                    * unit.angstrom)
    simulation.reporters.append(
        reporter.NetCDFReporter("traj_subset.nc", 1, periodic=True, velocities=True,
                                forces=True, subset=[0])
    )
    simulation.step(1)

    ncdf = file.NetCDFFile("traj_subset.nc", "r")
    assert ncdf.get_num_atoms() == 1

    # TEST CASE 7: Correct number of atoms and lack of velocities and
    # forces for full trajectory file
    state = simulation.context.getState(getPositions=True)
    file.NetCDFFile.write_model(
        "traj_two.nc",
        state.getTime().value_in_unit(unit.picosecond),
        state.getPositions(asNumpy=True).value_in_unit(unit.angstrom)
    )
    ncdf = file.NetCDFFile("traj_two.nc", "r")
    assert ncdf.get_num_atoms() == 2
    with pytest.warns(UserWarning):
        assert ncdf.get_velocities() is None
    with pytest.warns(UserWarning):
        assert ncdf.get_forces() is None


"""
Molecular structure
===================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains algorithms for computing structural information
for a collection of particles.
"""

from typing import Union

import MDAnalysis as mda
import numpy as np


def center_of_mass(
    group: mda.AtomGroup = None,
    grouping: str = None,
    *,
    masses: Union[np.ndarray[float], list[np.ndarray[float]]] = None,
    positions: Union[np.ndarray[float], list[np.ndarray[float]]] = None,
    images: Union[np.ndarray[int], list[np.ndarray[int]]] = None,
    dimensions: np.ndarray[float] = None,
    n_groups: int = None,
    raw: bool = False,
) -> Union[
    np.ndarray[float], tuple[np.ndarray[float], np.ndarray[float], np.ndarray[float]]
]:
    r"""
    Computes the centers of mass :math:`\mathbf{R}_\mathrm{com}` for a
    collection of atoms.

    For a group of :math:`N` atoms with masses :math:`m_i` and positions
    :math:`\mathbf{r}_i`, the center of mass is defined as

    .. math::

       \mathbf{R}_\mathrm{com}=\frac{\sum_{i=1}^Nm_i
       \mathbf{r}_i}{\sum_{i=1}^Nm_i}

    .. note::

       This function supports a wide variety of inputs, depending on
       how the atom information is provided and what should be
       calculated.

       When an :class:`MDAnalysis.core.groups.AtomGroup` object is
       provided in `group`, the atom masses and positions are retrieved
       from it and do not need to be provided in `masses` and
       `positions`, respectively. If the :code:`AtomGroup` abides by the
       standard topological heirarchy, you can specify the desired
       grouping in `grouping` and the appropriate centers of mass will
       be calculated. Otherwise, if and only if the :code:`AtomGroup`
       contains equisized or identical groups corresponding to the
       desired grouping (i.e., the :code:`AtomGroup` has atoms that are
       or can be treated as nonbonded entities or topological groups
       with the same number of but not necessarily identical
       constituents), you can provide the total number of groups in
       `n_groups` and the atom masses and positions will be distributed
       accordingly.

       If the trajectory is not unwrapped, the number of periodic
       boundary crossings (and optionally, the system dimensions if they
       are not embedded in the :code:`AtomGroup`) can be provided in
       `images` (and `dimensions`).

       If the :code:`AtomGroup` does not have the correct structural
       information and the residues or segments do not contain the same
       number of atoms, the atom masses and positions can each be
       provided directly as a :class:`numpy.ndarray` or list in `masses`
       and `positions`, respectively. To calculate the overall center of
       mass, the array-like object holding the masses should be
       one-dimensional, while that containing the positions should be
       two-dimensional. To calculate centers of mass for multiple
       groups, the array-like object holding the masses should be
       two-dimensional (indices: group, atom), while that containing the
       positions should be three-dimensional (indices: group, atom,
       axis). When a list is used, the inner arrays do not have to be
       homogeneously shaped, thus allowing you to calculate the centers
       of mass for residues or segments with different numbers of atoms.

       You may also provide only one of the atom masses or positions, in
       which case the missing information will be retrieved from the
       :code:`AtomGroup` provided in `group`. This is generally not
       recommended since the shapes of the provided and retrieved arrays
       may be incompatible.

    Parameters
    ----------
    group : `MDAnalysis.AtomGroup`, optional
        Collection of atoms to compute the centers of mass for. If not
        provided, the atom masses and posititions must be provided
        directly in `masses` and `positions`, respectively.

    grouping : `str`, optional
        Determines which center of mass is calculated if atom masses and
        positions are retrieved from `group`.

        .. container::

           **Valid values**:

           * :code:`None`: Center of mass of all atoms in `group`.
           * :code:`"residues"`: Centers of mass for each residue or
             molecule in `group`.
           * :code:`"segments"`: Centers of mass for each segment or
             chain in `group`.

    masses : array-like, keyword-only, optional
        Atom masses.

        .. container::

           **Shape**:

           The general ungrouped shape is :math:`(N,)`.

           For equisized or identical groups, the :class:`numpy.ndarray`
           object should have shape

           * :math:`(N,)` for the overall center of mass,
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res})` for the residue
             centers of mass, where :math:`N_\mathrm{res}` is
             the number of residues, or
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg}` for the segment
             centers of mass, where :math:`N_\mathrm{seg}` is
             the number of segments.

           For groups with different numbers of atoms, the list should
           contain inner array-like objects holding the masses of the
           atoms in each group.

        **Reference unit**: :math:`\mathrm{g/mol}`.

    positions : array-like, keyword-only, optional
        Atom positions.

        .. container::

           **Shape**:

           The general ungrouped shape is :math:`(N,\,3)`.

           For equisized or identical groups, the :class:`numpy.ndarray`
           object should have shape

           * :math:`(N,\,3)` for the overall center of mass,
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res},\,3)` for the
             residue centers of mass, or
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg},\,3)` for the
             segment centers of mass.

           For groups with different numbers of atoms, the list should
           contain inner array-like objects holding the positions of the
           atoms in each group.

        **Reference unit**: :math:`\mathrm{Å}`.

    images : array-like, keyword-only, optional
        Image flags for the atoms. Must be provided to get correct
        results if the trajectory is wrapped.

        **Shape**: Same as `positions`.

    dimensions : `numpy.ndarray`, keyword-only, optional
        System dimensions. Must be provided if `images` is provided and
        `group` is not provided or does not contain the system
        dimensions.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    n_groups : `int`, keyword-only, optional
        Number of residues or segments. Must be provided if `group` has
        an irregular topological heirarchy or the `masses` and
        `positions` arrays have the general ungrouped shapes.

    raw : `bool`, keyword-only, default: :code:`False`
        Determines whether atom masses and positions are returned.

    Returns
    -------
    com : `numpy.ndarray`
        Centers of mass.

        .. container::

           **Shape**:

           * :math:`(3,)` for :code:`grouping=None`.
           * :math:`(N_\mathrm{res},\,3)` for
             :code:`grouping="residues"`.
           * :math:`(N_\mathrm{seg},\,3)` for
             :code:`grouping="segments"`.

    masses : `numpy.ndarray`
        Atom masses. Only returned if `group` was provided and contains
        equisized or identical groups, and :code:`raw=True`.

        .. container::

           **Shape**:

           * :math:`(N,)` for :code:`grouping=None`.
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res})` for
             :code:`grouping="residues"`.
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg})` for
             :code:`grouping="segments"`.

        **Reference unit**: :math:`\mathrm{g/mol}`.

    positions : `numpy.ndarray`
        Unwrapped atom positions. Only returned if `group` was
        provided and contains equisized or identical groups, and
        :code:`raw=True`.

        .. container::

           **Shape**:

           * :math:`(N,\,3)` for :code:`grouping=None`.
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res},\,3)` for
             :code:`grouping="residues"`.
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg},\,3)` for
             :code:`grouping="segments"`.

        **Reference unit**: :math:`\mathrm{Å}`.

    Examples
    --------
    For an :class:`MDAnalysis.core.groups.AtomGroup` object with all
    necessary topological information and an unwrapped trajectory, the
    overall, per-residue, and per-segment centers of mass can be
    calculated as follows:

    >>> universe = mda.Universe("topology.pdb", "trajectory.nc")
    >>> group = universe.select_atoms("all")
    >>> com_ovr = center_of_mass(group)
    >>> com_res = center_of_mass(group, "residues")
    >>> com_seg = center_of_mass(group, "segments")

    If the :code:`AtomGroup` does not contain residue or segment
    information and the per-residue or per-segment centers of mass,
    respectively, are desired, the number of residues or segments can be
    provided in `n_groups`:

    >>> com_res = center_of_mass(group, "residues", n_groups=2) # 2 residues

    If the trajectory is wrapped, the number of periodic boundary
    crossings must be provided in `images`. Additionally, if the
    system dimensions are not embedded in the :code:`AtomGroup`, they
    must also be provided in `dimensions`:

    >>> images = np.array(((1, 0, 0), (0, 0, 0), (0, -1, 0),
    ...                    (0, 1, 0), (0, 0, 1), (-1, 1, 0)))
    >>> dimensions = np.array((5.0, 8.0, 10.0))
    >>> com_ovr = center_of_mass(group, images=images, dimensions=dimensions)

    Alternatively, if the atom masses and positions are directly
    available as :class:`numpy.ndarray` objects, the overall center of
    mass can be calculated as follows:

    >>> masses = np.array((12.01, 1.01, 1.01, 12.01, 1.01, 1.01))
    >>> positions = np.array(((0.0, -0.07579, 0.0),
    ...                       (0.86681, 0.60144, 0.0),
    ...                       (-0.86681, 0.60144, 0.0),
    ...                       (0.0, -0.07579, 1.0),
    ...                       (0.86681, 0.60144, 1.0),
    ...                       (-0.86681, 0.60144, 1.0)))
    >>> com_ovr = center_of_mass(masses=masses, positions=positions)

    If the per-residue center of mass is desired, the number of residues
    can be provided in `n_groups`:

    >>> com_res = center_of_mass(masses=masses, positions=positions, n_groups=2)

    or the arrays containing the atom masses and positions can be
    reshaped to the appropriate shapes:

    >>> n_groups = 2
    >>> com_res = center_of_mass(masses=masses.reshape((n_groups, -1)),
    ...                          positions=positions.reshape((n_groups, -1, 3)))

    Like before, if the trajectory is wrapped, the number of periodic
    boundary crossings and system dimensions must be provided in
    `images` and `dimensions`, respectively:

    >>> images = np.array(((0, 0, 0), (0, 0, 0), (0, 0, 0),
    ...                    (0, 0, 1), (0, 0, 1), (0, 0, 1)))
    >>> dimensions = np.array((12.0, 12.0, 12.0))
    >>> com_ovr = center_of_mass(masses=masses, positions=positions,
    ...                          images=images, dimensions=dimensions)

    Finally, if the per-residue or per-segment center of mass is
    desired but the groups contain different numbers of atoms, the
    the atom masses and positions can be provided as lists of arrays
    holding the masses and positions of the atoms in each group:

    >>> masses = [(12.01, 1.01, 1.01), (22.99,), (12.01, 1.01)]
    >>> positions = [((0.0, -0.07579, 0.0),
    ...               (0.86681, 0.60144, 0.0),
    ...               (-0.86681, 0.60144, 0.0)),
    ...              ((0.0, 0.0, 0.0),),
    ...              ((0.0, -0.07579, 1.0),
    ...               (0.86681, 0.60144, 1.0))]
    >>> com_res = center_of_mass(masses=masses, positions=positions)

    It is still possible to pass in the number of periodic boundary
    crossings if the trajectory is wrapped, but attention must be paid
    to the shape of the array:

    >>> images = [((0, 0, 0), (0, 0, 0), (0, 0, 0)),
    ...           ((1, 1, 1),),
    ...           ((0, 1, 0), (0, 1, 0))]
    >>> dimensions = np.array((10.0, 10.0, 10.0))
    >>> com_res = center_of_mass(masses=masses, positions=positions,
    ...                          images=images, dimensions=dimensions)
    """

    # Check whether grouping is valid
    if grouping not in {None, "residues", "segments"}:
        emsg = (
            f"Invalid grouping: '{grouping}'. Valid options are "
            "None, 'residues', and 'segments'."
        )
        raise ValueError(emsg)

    # Get system dimensions if image flags are provided
    if images is not None:
        if dimensions is None:
            try:
                dimensions = group.dimensions[:3]
            except (NameError, TypeError):
                emsg = (
                    "Image flags were provided, but no system "
                    "dimensions were provided or found in the "
                    "trajectory."
                )
                raise ValueError(emsg)
        else:
            dimensions = np.asarray(dimensions)

    # Get particle masses and positions from the trajectory, if
    # necessary
    missing = (masses is None, positions is None)
    if any(missing):
        if group is None:
            emsg = (
                "Either a group of atoms or atom positions and "
                "masses must be provided."
            )
            raise ValueError(emsg)

        # Check whether the groups have equal numbers of atoms
        if grouping is None:
            same = True
        else:
            groups = getattr(group, grouping)

            # Calculate and return the centers of mass for different
            # groups here if unwrapping and the mass and position arrays
            # are not needed
            if (
                not (
                    same := all(
                        g.atoms.n_atoms == groups[0].atoms.n_atoms for g in groups
                    )
                )
                and images is None
                and not raw
            ):
                return np.array([g.atoms.center_of_mass() for g in groups])

        # Get and unwrap particle positions, if necessary
        if missing[1]:
            positions = group.positions
            if images is not None:
                positions += images * dimensions[:3]

        # Get particle masses and ensure correct dimensionality, if
        # necessary
        if same:
            if missing[0]:
                masses = group.masses
            if grouping is not None or n_groups is not None:
                shape = (n_groups or getattr(group, f"n_{grouping}"), -1, 3)
                masses = masses.reshape(shape[:-1])
                positions = positions.reshape(shape)
        else:
            if missing[0]:
                masses = [g.atoms.masses for g in groups]
            if missing[1]:
                positions = [positions[g.atoms.ix] for g in groups]
    else:

        # Try to convert arrays to NumPy arrays if they are not already
        # to take advantage of vectorized operations later
        try:
            positions = np.asarray(positions)
            masses = np.asarray(masses)
            if images is not None:
                positions += images
        except ValueError:
            pass
        if type(masses) is not type(positions):
            emsg = (
                "The shapes of the arrays containing the masses "
                "and positions are incompatible."
            )
            raise ValueError(emsg)
        if images is not None and type(images) is not type(positions):
            emsg = (
                "The shapes of the arrays containing the positions "
                "and image flags are incompatible."
            )
            raise ValueError(emsg)

    # Calculate the centers of mass for the specified grouping
    if isinstance(positions, np.ndarray):

        # Reshape the mass and position arrays based on the specified
        # number of groups
        if n_groups is not None:
            masses = masses.reshape((n_groups, -1))
            positions = positions.reshape((n_groups, -1, 3))

        com = np.einsum("...a,...ad->...d", masses, positions) / masses.sum(
            axis=-1, keepdims=True
        )
    else:
        if images is not None:
            for j, (p, i) in enumerate(zip(positions, images)):
                positions[j] = p + i * dimensions
        com = np.array([np.dot(m, p) / m.sum() for m, p in zip(masses, positions)])

    # Return raw masses and positions, if desired
    if raw:
        return com, masses, positions

    return com


def radius_of_gyration(
    group: mda.AtomGroup = None,
    grouping: str = None,
    *,
    masses: Union[np.ndarray[float], list[np.ndarray[float]]] = None,
    positions: Union[np.ndarray[float], list[np.ndarray[float]]] = None,
    com: np.ndarray[float] = None,
    images: Union[np.ndarray[int], list[np.ndarray[int]]] = None,
    dimensions: np.ndarray[float] = None,
    n_groups: int = None,
    components: bool = False,
) -> Union[float, np.ndarray[float]]:
    r"""
    Computes the radii of gyration :math:`R_\mathrm{g}` for a collection
    of atoms.

    For a group of :math:`N` atoms with masses :math:`m_i` and positions
    :math:`\mathbf{r}_i`, the radius of gyration is defined as

    .. math::

       R_\mathrm{g}=\sqrt{
       \frac{\sum_i^Nm_i\|\mathbf{r}_i-\mathbf{R}_\mathrm{com}\|^2}
       {\sum_i^Nm_i}}

    where :math:`\mathbf{R}_\mathrm{com}` is the center of mass.

    Alternatively, the radii of gyration around the coordinate axes can
    be calculated by only summing the radii components orthogonal to
    each axis. For example, the radius of gyration around the
    :math:`x`-axis is

    .. math::

       R_{\mathrm{g},\,x}=\sqrt{\frac{\sum_i^Nm_i
       \left[(\mathbf{r}_{i,\,y}-\mathbf{R}_{\mathrm{com},\,y})^2
       +(\mathbf{r}_{i,\,z}-\mathbf{R}_{\mathrm{com},\,z})^2\right]}
       {\sum_i^Nm_i}}

    .. note::

       This function supports a wide variety of inputs, depending on
       how the atom information is provided and what should be
       calculated.

       When an :class:`MDAnalysis.core.groups.AtomGroup` object is
       provided in `group`, the atom masses and positions are retrieved
       from it and do not need to be provided in `masses` and
       `positions`, respectively. If the :code:`AtomGroup` abides by the
       standard topological heirarchy, you can specify the desired
       grouping in `grouping` and the appropriate radii of gyration will
       be calculated. Otherwise, if and only if the :code:`AtomGroup`
       contains equisized or identical groups corresponding to the
       desired grouping (i.e., the :code:`AtomGroup` has atoms that are
       or can be treated as nonbonded entities or topological groups
       with the same number of but not necessarily identical
       constituents), you can provide the total number of groups in
       `n_groups` and the atom masses and positions will be distributed
       accordingly.

       If the trajectory is not unwrapped, the number of periodic
       boundary crossings (and optionally, the system dimensions if they
       are not embedded in the :code:`AtomGroup`) can be provided in
       `images` (and `dimensions`).

       If the :code:`AtomGroup` does not have the correct structural
       information and the residues or segments do not contain the same
       number of atoms, the atom masses and positions can each be
       provided directly as a :class:`numpy.ndarray` or list in `masses`
       and `positions`, respectively. To calculate the overall radius of
       gyration, the array-like object holding the masses should be
       one-dimensional, while that containing the positions should be
       two-dimensional. To calculate radii of gyration for multiple
       groups, the array-like object holding the masses should be
       two-dimensional (indices: group, atom), while that containing the
       positions should be three-dimensional (indices: group, atom,
       axis). When a list is used, the inner arrays do not have to be
       homogeneously shaped, thus allowing you to calculate the radii of
       gyration for residues or segments with different numbers of atoms.

       You may also provide only one of the atom masses or positions, in
       which case the missing information will be retrieved from the
       :code:`AtomGroup` provided in `group`. This is generally not
       recommended since the shapes of the provided and retrieved arrays
       may be incompatible.

    Parameters
    ----------
    group : `MDAnalysis.AtomGroup`, optional
        Collection of atoms to compute the radii of gyration for. If not
        provided, the atom masses and posititions must be provided
        directly in `masses` and `positions`, respectively.

    grouping : `str`, optional
        Determines which radius of gyration is calculated if atom masses
        and positions are retrieved from `group`.

        .. container::

           **Valid values**:

           * :code:`None`: Radius of gyration of all atoms in `group`.
           * :code:`"residues"`: Radius of gyration for each residue or
             molecule in `group`.
           * :code:`"segments"`: Radius of gyration for each segment or
             chain in `group`.

    masses : array-like, keyword-only, optional
        Atom masses.

        .. container::

           **Shape**:

           The general ungrouped shape is :math:`(N,)`.

           For equisized or identical groups, the :class:`numpy.ndarray`
           object should have shape

           * :math:`(N,)` for the overall radius of gyration,
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res})` for the residue
             radii of gyration, where :math:`N_\mathrm{res}` is the
             number of residues, or
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg}` for the segment
             radii of gyration, where :math:`N_\mathrm{seg}` is the
             number of segments.

           For groups with different numbers of atoms, the list should
           contain inner array-like objects holding the masses of the
           atoms in each group.

        **Reference unit**: :math:`\mathrm{g/mol}`.

    positions : array-like, keyword-only, optional
        Atom positions.

        .. container::

           **Shape**:

           The general ungrouped shape is :math:`(N,\,3)`.

           For equisized or identical groups, the :class:`numpy.ndarray`
           object should have shape

           * :math:`(N,\,3)` for the overall radius of gyration,
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res},\,3)` for the
             residue radii of gyration, or
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg},\,3)` for the
             segment radii of gyration.

           For groups with different numbers of atoms, the list should
           contain inner array-like objects holding the positions of the
           atoms in each group.

        **Reference unit**: :math:`\mathrm{Å}`.

    com : `numpy.ndarray`, keyword-only, optional
        Centers of mass.

        .. container::

           **Shape**:

           * :math:`(3,)` for the overall radius of gyration.
           * :math:`(N_\mathrm{res},\,3)` for the residue radii of
             gyration.
           * :math:`(N_\mathrm{seg},\,3)` for the segment radii of
             gyration.

    images : array-like, keyword-only, optional
        Image flags for the atoms. Must be provided to get correct
        results if the trajectory is wrapped.

        **Shape**: Same as `positions`.

    dimensions : `numpy.ndarray`, keyword-only, optional
        System dimensions. Must be provided if `images` is provided and
        `group` is not provided or does not contain the system
        dimensions.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    n_groups : `int`, keyword-only, optional
        Number of residues or segments. Must be provided if `group` has
        an irregular topological heirarchy or the `masses` and
        `positions` arrays have the general ungrouped shapes.

    components : `bool`, keyword-only, default: :code:`False`
        Specifies whether the components of the radii of gyration are
        calculated and returned instead.

    Returns
    -------
    gyradii : `float` or `numpy.ndarray`
        Radii of gyration.

        .. container::

           **Shape**:

           * Scalar for :code:`grouping=None`.
           * :math:`(N_\mathrm{res},)` for :code:`grouping="residues"`.
           * :math:`(N_\mathrm{seg},)` for :code:`grouping="segments"`.

           If :code:`components=True`, an additional axis with length
           :math:`3` is added to the end.

        **Reference unit**: :math:`\mathrm{Å}`.

    Examples
    --------
    For an :class:`MDAnalysis.core.groups.AtomGroup` object with all
    necessary topological information and an unwrapped trajectory, the
    overall, per-residue, and per-segment radii of gyration can be
    calculated as follows:

    >>> universe = mda.Universe("topology.pdb", "trajectory.nc")
    >>> group = universe.select_atoms("all")
    >>> Rg_ovr = radius_of_gyration(group)
    >>> Rg_res = radius_of_gyration(group, "residues")
    >>> Rg_seg = radius_of_gyration(group, "segments")

    If the :code:`AtomGroup` does not contain residue or segment
    information and the per-residue or per-segment centers of mass,
    respectively, are desired, the number of residues or segments can be
    provided in `n_groups`:

    >>> Rg_res = radius_of_gyration(group, "residues", n_groups=2) # 2 residues

    If the trajectory is wrapped, the number of periodic boundary
    crossings must be provided in `images`. Additionally, if the
    system dimensions are not embedded in the :code:`AtomGroup`, they
    must also be provided in `dimensions`:

    >>> images = np.array(((1, 0, 0), (0, 0, 0), (0, -1, 0),
    ...                    (0, 1, 0), (0, 0, 1), (-1, 1, 0)))
    >>> dimensions = np.array((5.0, 8.0, 10.0))
    >>> Rg_ovr = radius_of_gyration(group, images=images, dimensions=dimensions)

    Alternatively, if the atom masses and positions are directly
    available as :class:`numpy.ndarray` objects, the overall radius of
    gyration can be calculated as follows:

    >>> masses = np.array((12.01, 1.01, 1.01, 12.01, 1.01, 1.01))
    >>> positions = np.array(((0.0, -0.07579, 0.0),
    ...                       (0.86681, 0.60144, 0.0),
    ...                       (-0.86681, 0.60144, 0.0),
    ...                       (0.0, -0.07579, 1.0),
    ...                       (0.86681, 0.60144, 1.0),
    ...                       (-0.86681, 0.60144, 1.0)))
    >>> Rg_ovr = radius_of_gyration(masses=masses, positions=positions)

    If the per-residue radius of gyration is desired, the number of
    residues can be provided in `n_groups`:

    >>> Rg_res = radius_of_gyration(masses=masses, positions=positions,
    ...                             n_groups=2)

    or the arrays containing the atom masses and positions can be
    reshaped to the appropriate shapes:

    >>> n_groups = 2
    >>> Rg_res = radius_of_gyration(
    ...     masses=masses.reshape((n_groups, -1)),
    ...     positions=positions.reshape((n_groups, -1, 3))
    ... )

    Like before, if the trajectory is wrapped, the number of periodic
    boundary crossings and system dimensions must be provided in
    `images` and `dimensions`, respectively:

    >>> images = np.array(((0, 0, 0), (0, 0, 0), (0, 0, 0),
    ...                    (0, 0, 1), (0, 0, 1), (0, 0, 1)))
    >>> dimensions = np.array((12.0, 12.0, 12.0))
    >>> Rg_ovr = radius_of_gyration(masses=masses, positions=positions,
    ...                             images=images, dimensions=dimensions)

    Finally, if the per-residue or per-segment radius of gyration is
    desired but the groups contain different numbers of atoms, the
    the atom masses and positions can be provided as lists of arrays
    holding the masses and positions of the atoms in each group:

    >>> masses = [(12.01, 1.01, 1.01), (22.99,), (12.01, 1.01)]
    >>> positions = [((0.0, -0.07579, 0.0),
    ...               (0.86681, 0.60144, 0.0),
    ...               (-0.86681, 0.60144, 0.0)),
    ...              ((0.0, 0.0, 0.0),),
    ...              ((0.0, -0.07579, 1.0),
    ...               (0.86681, 0.60144, 1.0))]
    >>> Rg_res = radius_of_gyration(masses=masses, positions=positions)

    It is still possible to pass in the number of periodic boundary
    crossings if the trajectory is wrapped, but attention must be paid
    to the shape of the array:

    >>> images = [((0, 0, 0), (0, 0, 0), (0, 0, 0)),
    ...           ((1, 1, 1),),
    ...           ((0, 1, 0), (0, 1, 0))]
    >>> dimensions = np.array((10.0, 10.0, 10.0))
    >>> Rg_res = radius_of_gyration(masses=masses, positions=positions,
    ...                             images=images, dimensions=dimensions)

    For any of the examples above, the components of the radii of
    gyration can be calculated and returned by setting
    :code:`components=True`:

    >>> Rg_ovr = radius_of_gyration(group, components=True)
    """

    # Check whether grouping is valid
    if grouping not in {None, "residues", "segments"}:
        emsg = (
            f"Invalid grouping: '{grouping}'. Valid options are "
            "None, 'residues', and 'segments'."
        )
        raise ValueError(emsg)

    # Get particle masses and positions from the trajectory and the
    # center(s) of mass, if necessary
    missing = (masses is None, positions is None, com is None)
    if any(missing[:2]):
        com, masses, positions = center_of_mass(
            group,
            grouping,
            masses=masses,
            positions=positions,
            images=images,
            dimensions=dimensions,
            n_groups=n_groups,
            raw=True,
        )
    elif missing[2]:
        com, masses, positions = center_of_mass(
            masses=masses,
            positions=positions,
            images=images,
            dimensions=dimensions,
            n_groups=n_groups,
            raw=True,
        )

    if isinstance(positions, np.ndarray):
        if components:
            cpos = (positions - np.expand_dims(com, axis=positions.ndim - 2)) ** 2

            # Compute the radii of gyration in each direction for
            # equisized or identical groups
            if grouping or n_groups:
                return np.sqrt(
                    np.einsum(
                        "ga,gad->gd",
                        masses,
                        np.stack(
                            (
                                cpos[:, :, (1, 2)].sum(axis=2),
                                cpos[:, :, (0, 2)].sum(axis=2),
                                cpos[:, :, (0, 1)].sum(axis=2),
                            ),
                            axis=2,
                        ),
                    )
                    / masses.sum(axis=1, keepdims=True)
                )

            # Compute the radius of gyration in each direction for all
            # atoms
            return np.sqrt(
                np.dot(
                    masses,
                    np.hstack(
                        (
                            cpos[:, (1, 2)].sum(axis=1, keepdims=True),
                            cpos[:, (0, 2)].sum(axis=1, keepdims=True),
                            cpos[:, (0, 1)].sum(axis=1, keepdims=True),
                        )
                    ),
                )
                / masses.sum()
            )

        # Compute the overall radii of gyration for equisized or
        # identical groups
        elif grouping or n_groups:
            return np.sqrt(
                np.einsum("ga,gad->gd", masses, (positions - com[:, None]) ** 2).sum(
                    axis=1
                )
                / masses.sum(axis=1)
            )

        # Compute the overall radius of gyration for all atoms
        return np.sqrt(np.dot(masses, (positions - com) ** 2).sum() / masses.sum())

    # Compute the radii of gyration in each direction for asymmetric
    # groups
    if components:
        gyradii = np.empty(com.shape)
        for i, (m, p, c) in enumerate(zip(masses, positions, com)):
            cpos = (p - c) ** 2
            gyradii[i] = (
                np.array(
                    (
                        np.dot(m, cpos[:, (1, 2)].sum(axis=1)),
                        np.dot(m, cpos[:, (0, 2)].sum(axis=1)),
                        np.dot(m, cpos[:, (0, 1)].sum(axis=1)),
                    )
                )
                / m.sum()
            )
        return np.sqrt(gyradii)

    # Compute the overall radii of gyration for asymmetric groups
    return np.sqrt(
        [
            np.einsum("a,ad->d", m, (p - c) ** 2).sum() / m.sum()
            for m, p, c in zip(masses, positions, com)
        ]
    )



import pathlib
import sys

import numpy as np
import pytest

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.algorithm import utility # noqa: E402

rng = np.random.default_rng()

def test_func_closest_factors():

    # TEST CASE 1: Cube root of perfect cube
    factors = utility.get_closest_factors(1000, 3)
    assert np.allclose(factors, 10 * np.ones(3, dtype=int))

    # TEST CASE 2: Three closest factors in ascending order
    factors = utility.get_closest_factors(35904, 3)
    assert factors.tolist() == [32, 33, 34]

    # TEST CASE 3: Four closest factors in descending order
    factors = utility.get_closest_factors(73440, 4, reverse=True)
    assert factors.tolist() == [18, 17, 16, 15]

def test_func_replicate():

    # TEST CASE 1: Replicate two vectors
    dims = rng.integers(1, 5, size=3)
    n_cells = rng.integers(2, 10, size=3)
    pos = utility.replicate(dims, np.array(((0, 0, 0), dims // 2)), n_cells)
    assert pos.shape[0] == 2 * n_cells.prod()
    assert np.allclose(pos[2], (dims[0], 0, 0))

def test_func_rebin():

    # TEST CASE 1: Rebin 1D array
    arr = np.arange(50)
    ref = np.arange(2, 52, 5)
    assert np.allclose(utility.rebin(arr), ref)

    # TEST CASE 2: Rebin 2D array
    assert np.allclose(utility.rebin(np.tile(arr[None, :], (5, 1))),
                       np.tile(ref[None, :], (5, 1)))

    # TEST CASE 3: No factor specified and cannot be determined
    with pytest.raises(ValueError):
        utility.rebin(np.empty((17,)))